#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# flake8: noqa


# This file is automatically generated. Please do not modify it directly.
# Find the relevant recipe file in the samples/recipes or samples/ingredients
# directory and apply your changes there.


# [START compute_custom_machine_type_create_with_helper]
from __future__ import annotations

from collections import namedtuple
from enum import Enum
from enum import unique
import re
import sys
from typing import Any
import warnings

from google.api_core.extended_operation import ExtendedOperation
from google.cloud import compute_v1


def gb_to_mb(value: int) -> int:
    return value << 10


class CustomMachineType:
    """
    Allows to create custom machine types to be used with the VM instances.
    """

    @unique
    class CPUSeries(Enum):
        N1 = "custom"
        N2 = "n2-custom"
        N2D = "n2d-custom"
        E2 = "e2-custom"
        E2_MICRO = "e2-custom-micro"
        E2_SMALL = "e2-custom-small"
        E2_MEDIUM = "e2-custom-medium"

    TypeLimits = namedtuple(
        "TypeLimits",
        [
            "allowed_cores",
            "min_mem_per_core",
            "max_mem_per_core",
            "allow_extra_memory",
            "extra_memory_limit",
        ],
    )

    # The limits for various CPU types are described on:
    # https://cloud.google.com/compute/docs/general-purpose-machines
    LIMITS = {
        CPUSeries.E2: TypeLimits(frozenset(range(2, 33, 2)), 512, 8192, False, 0),
        CPUSeries.E2_MICRO: TypeLimits(frozenset(), 1024, 2048, False, 0),
        CPUSeries.E2_SMALL: TypeLimits(frozenset(), 2048, 4096, False, 0),
        CPUSeries.E2_MEDIUM: TypeLimits(frozenset(), 4096, 8192, False, 0),
        CPUSeries.N2: TypeLimits(
            frozenset(range(2, 33, 2)).union(set(range(36, 129, 4))),
            512,
            8192,
            True,
            gb_to_mb(624),
        ),
        CPUSeries.N2D: TypeLimits(
            frozenset({2, 4, 8, 16, 32, 48, 64, 80, 96}), 512, 8192, True, gb_to_mb(768)
        ),
        CPUSeries.N1: TypeLimits(
            frozenset({1}.union(range(2, 97, 2))), 922, 6656, True, gb_to_mb(624)
        ),
    }

    def __init__(
        self, zone: str, cpu_series: CPUSeries, memory_mb: int, core_count: int = 0
    ):
        self.zone = zone
        self.cpu_series = cpu_series
        self.limits = self.LIMITS[self.cpu_series]
        # Shared machine types (e2-small, e2-medium and e2-micro) always have
        # 2 vCPUs: https://cloud.google.com/compute/docs/general-purpose-machines#e2_limitations
        self.core_count = 2 if self.is_shared() else core_count
        self.memory_mb = memory_mb
        self._checked = False
        self._check_parameters()
        self.extra_memory_used = self._check_extra_memory()

    def is_shared(self):
        return self.cpu_series in (
            CustomMachineType.CPUSeries.E2_SMALL,
            CustomMachineType.CPUSeries.E2_MICRO,
            CustomMachineType.CPUSeries.E2_MEDIUM,
        )

    def _check_extra_memory(self) -> bool:
        if self._checked:
            return self.memory_mb > self.core_count * self.limits.max_mem_per_core
        else:
            raise RuntimeError(
                "You need to call _check_parameters() before calling _check_extra_memory()"
            )

    def _check_parameters(self):
        """
        Check whether the requested parameters are allowed. Find more information about limitations of custom machine
        types at: https://cloud.google.com/compute/docs/general-purpose-machines#custom_machine_types
        """
        # Check the number of cores
        if (
            self.limits.allowed_cores
            and self.core_count not in self.limits.allowed_cores
        ):
            raise RuntimeError(
                f"Invalid number of cores requested. Allowed number of cores for {self.cpu_series.name} is: {sorted(self.limits.allowed_cores)}"
            )

        # Memory must be a multiple of 256 MB
        if self.memory_mb % 256 != 0:
            raise RuntimeError("Requested memory must be a multiple of 256 MB.")

        # Check if the requested memory isn't too little
        if self.memory_mb < self.core_count * self.limits.min_mem_per_core:
            raise RuntimeError(
                f"Requested memory is too low. Minimal memory for {self.cpu_series.name} is {self.limits.min_mem_per_core} MB per core."
            )

        # Check if the requested memory isn't too much
        if self.memory_mb > self.core_count * self.limits.max_mem_per_core:
            if self.limits.allow_extra_memory:
                if self.memory_mb > self.limits.extra_memory_limit:
                    raise RuntimeError(
                        f"Requested memory is too large.. Maximum memory allowed for {self.cpu_series.name} is {self.limits.extra_memory_limit} MB."
                    )
            else:
                raise RuntimeError(
                    f"Requested memory is too large.. Maximum memory allowed for {self.cpu_series.name} is {self.limits.max_mem_per_core} MB per core."
                )

        self._checked = True

    def __str__(self) -> str:
        """
        Return the custom machine type in form of a string acceptable by Compute Engine API.
        """
        if self.cpu_series in {
            self.CPUSeries.E2_SMALL,
            self.CPUSeries.E2_MICRO,
            self.CPUSeries.E2_MEDIUM,
        }:
            return f"zones/{self.zone}/machineTypes/{self.cpu_series.value}-{self.memory_mb}"

        if self.extra_memory_used:
            return f"zones/{self.zone}/machineTypes/{self.cpu_series.value}-{self.core_count}-{self.memory_mb}-ext"

        return f"zones/{self.zone}/machineTypes/{self.cpu_series.value}-{self.core_count}-{self.memory_mb}"

    def short_type_str(self) -> str:
        """
        Return machine type in a format without the zone. For example, n2-custom-0-10240.
        This format is used to create instance templates.
        """
        return str(self).rsplit("/", maxsplit=1)[1]

    @classmethod
    def from_str(cls, machine_type: str):
        """
        Construct a new object from a string. The string needs to be a valid custom machine type like:
         - https://www.googleapis.com/compute/v1/projects/diregapic-mestiv/zones/us-central1-b/machineTypes/e2-custom-4-8192
         - zones/us-central1-b/machineTypes/e2-custom-4-8192
         - e2-custom-4-8192 (in this case, the zone parameter will not be set)
        """
        zone = None
        if machine_type.startswith("http"):
            machine_type = machine_type[machine_type.find("zones/") :]

        if machine_type.startswith("zones/"):
            _, zone, _, machine_type = machine_type.split("/")

        extra_mem = machine_type.endswith("-ext")

        if machine_type.startswith("custom"):
            cpu = cls.CPUSeries.N1
            _, cores, memory = machine_type.rsplit("-", maxsplit=2)
        else:
            if extra_mem:
                cpu_series, _, cores, memory, _ = machine_type.split("-")
            else:
                cpu_series, _, cores, memory = machine_type.split("-")
            if cpu_series == "n2":
                cpu = cls.CPUSeries.N2
            elif cpu_series == "n2d":
                cpu = cls.CPUSeries.N2D
            elif cpu_series == "e2":
                cpu = cls.CPUSeries.E2
                if cores == "micro":
                    cpu = cls.CPUSeries.E2_MICRO
                    cores = 2
                elif cores == "small":
                    cpu = cls.CPUSeries.E2_SMALL
                    cores = 2
                elif cores == "medium":
                    cpu = cls.CPUSeries.E2_MEDIUM
                    cores = 2
            else:
                raise RuntimeError("Unknown CPU series.")

        cores = int(cores)
        memory = int(memory)

        return cls(zone, cpu, memory, cores)


def get_image_from_family(project: str, family: str) -> compute_v1.Image:
    """
    Retrieve the newest image that is part of a given family in a project.

    Args:
        project: project ID or project number of the Cloud project you want to get image from.
        family: name of the image family you want to get image from.

    Returns:
        An Image object.
    """
    image_client = compute_v1.ImagesClient()
    # List of public operating system (OS) images: https://cloud.google.com/compute/docs/images/os-details
    newest_image = image_client.get_from_family(project=project, family=family)
    return newest_image


def disk_from_image(
    disk_type: str,
    disk_size_gb: int,
    boot: bool,
    source_image: str,
    auto_delete: bool = True,
) -> compute_v1.AttachedDisk:
    """
    Create an AttachedDisk object to be used in VM instance creation. Uses an image as the
    source for the new disk.

    Args:
         disk_type: the type of disk you want to create. This value uses the following format:
            "zones/{zone}/diskTypes/(pd-standard|pd-ssd|pd-balanced|pd-extreme)".
            For example: "zones/us-west3-b/diskTypes/pd-ssd"
        disk_size_gb: size of the new disk in gigabytes
        boot: boolean flag indicating whether this disk should be used as a boot disk of an instance
        source_image: source image to use when creating this disk. You must have read access to this disk. This can be one
            of the publicly available images or an image from one of your projects.
            This value uses the following format: "projects/{project_name}/global/images/{image_name}"
        auto_delete: boolean flag indicating whether this disk should be deleted with the VM that uses it

    Returns:
        AttachedDisk object configured to be created using the specified image.
    """
    boot_disk = compute_v1.AttachedDisk()
    initialize_params = compute_v1.AttachedDiskInitializeParams()
    initialize_params.source_image = source_image
    initialize_params.disk_size_gb = disk_size_gb
    initialize_params.disk_type = disk_type
    boot_disk.initialize_params = initialize_params
    # Remember to set auto_delete to True if you want the disk to be deleted when you delete
    # your VM instance.
    boot_disk.auto_delete = auto_delete
    boot_disk.boot = boot
    return boot_disk


def wait_for_extended_operation(
    operation: ExtendedOperation, verbose_name: str = "operation", timeout: int = 300
) -> Any:
    """
    Waits for the extended (long-running) operation to complete.

    If the operation is successful, it will return its result.
    If the operation ends with an error, an exception will be raised.
    If there were any warnings during the execution of the operation
    they will be printed to sys.stderr.

    Args:
        operation: a long-running operation you want to wait on.
        verbose_name: (optional) a more verbose name of the operation,
            used only during error and warning reporting.
        timeout: how long (in seconds) to wait for operation to finish.
            If None, wait indefinitely.

    Returns:
        Whatever the operation.result() returns.

    Raises:
        This method will raise the exception received from `operation.exception()`
        or RuntimeError if there is no exception set, but there is an `error_code`
        set for the `operation`.

        In case of an operation taking longer than `timeout` seconds to complete,
        a `concurrent.futures.TimeoutError` will be raised.
    """
    result = operation.result(timeout=timeout)

    if operation.error_code:
        print(
            f"Error during {verbose_name}: [Code: {operation.error_code}]: {operation.error_message}",
            file=sys.stderr,
            flush=True,
        )
        print(f"Operation ID: {operation.name}", file=sys.stderr, flush=True)
        raise operation.exception() or RuntimeError(operation.error_message)

    if operation.warnings:
        print(f"Warnings during {verbose_name}:\n", file=sys.stderr, flush=True)
        for warning in operation.warnings:
            print(f" - {warning.code}: {warning.message}", file=sys.stderr, flush=True)

    return result


def create_instance(
    project_id: str,
    zone: str,
    instance_name: str,
    disks: list[compute_v1.AttachedDisk],
    machine_type: str = "n1-standard-1",
    network_link: str = "global/networks/default",
    subnetwork_link: str = None,
    internal_ip: str = None,
    external_access: bool = False,
    external_ipv4: str = None,
    accelerators: list[compute_v1.AcceleratorConfig] = None,
    preemptible: bool = False,
    spot: bool = False,
    instance_termination_action: str = "STOP",
    custom_hostname: str = None,
    delete_protection: bool = False,
) -> compute_v1.Instance:
    """
    Send an instance creation request to the Compute Engine API and wait for it to complete.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        zone: name of the zone to create the instance in. For example: "us-west3-b"
        instance_name: name of the new virtual machine (VM) instance.
        disks: a list of compute_v1.AttachedDisk objects describing the disks
            you want to attach to your new instance.
        machine_type: machine type of the VM being created. This value uses the
            following format: "zones/{zone}/machineTypes/{type_name}".
            For example: "zones/europe-west3-c/machineTypes/f1-micro"
        network_link: name of the network you want the new instance to use.
            For example: "global/networks/default" represents the network
            named "default", which is created automatically for each project.
        subnetwork_link: name of the subnetwork you want the new instance to use.
            This value uses the following format:
            "regions/{region}/subnetworks/{subnetwork_name}"
        internal_ip: internal IP address you want to assign to the new instance.
            By default, a free address from the pool of available internal IP addresses of
            used subnet will be used.
        external_access: boolean flag indicating if the instance should have an external IPv4
            address assigned.
        external_ipv4: external IPv4 address to be assigned to this instance. If you specify
            an external IP address, it must live in the same region as the zone of the instance.
            This setting requires `external_access` to be set to True to work.
        accelerators: a list of AcceleratorConfig objects describing the accelerators that will
            be attached to the new instance.
        preemptible: boolean value indicating if the new instance should be preemptible
            or not. Preemptible VMs have been deprecated and you should now use Spot VMs.
        spot: boolean value indicating if the new instance should be a Spot VM or not.
        instance_termination_action: What action should be taken once a Spot VM is terminated.
            Possible values: "STOP", "DELETE"
        custom_hostname: Custom hostname of the new VM instance.
            Custom hostnames must conform to RFC 1035 requirements for valid hostnames.
        delete_protection: boolean value indicating if the new virtual machine should be
            protected against deletion or not.
    Returns:
        Instance object.
    """
    instance_client = compute_v1.InstancesClient()

    # Use the network interface provided in the network_link argument.
    network_interface = compute_v1.NetworkInterface()
    network_interface.name = network_link
    if subnetwork_link:
        network_interface.subnetwork = subnetwork_link

    if internal_ip:
        network_interface.network_i_p = internal_ip

    if external_access:
        access = compute_v1.AccessConfig()
        access.type_ = compute_v1.AccessConfig.Type.ONE_TO_ONE_NAT.name
        access.name = "External NAT"
        access.network_tier = access.NetworkTier.PREMIUM.name
        if external_ipv4:
            access.nat_i_p = external_ipv4
        network_interface.access_configs = [access]

    # Collect information into the Instance object.
    instance = compute_v1.Instance()
    instance.network_interfaces = [network_interface]
    instance.name = instance_name
    instance.disks = disks
    if re.match(r"^zones/[a-z\d\-]+/machineTypes/[a-z\d\-]+$", machine_type):
        instance.machine_type = machine_type
    else:
        instance.machine_type = f"zones/{zone}/machineTypes/{machine_type}"

    if accelerators:
        instance.guest_accelerators = accelerators

    if preemptible:
        # Set the preemptible setting
        warnings.warn(
            "Preemptible VMs are being replaced by Spot VMs.", DeprecationWarning
        )
        instance.scheduling = compute_v1.Scheduling()
        instance.scheduling.preemptible = True

    if spot:
        # Set the Spot VM setting
        instance.scheduling = compute_v1.Scheduling()
        instance.scheduling.provisioning_model = (
            compute_v1.Scheduling.ProvisioningModel.SPOT.name
        )
        instance.scheduling.instance_termination_action = instance_termination_action

    if custom_hostname is not None:
        # Set the custom hostname for the instance
        instance.hostname = custom_hostname

    if delete_protection:
        # Set the delete protection bit
        instance.deletion_protection = True

    # Prepare the request to insert an instance.
    request = compute_v1.InsertInstanceRequest()
    request.zone = zone
    request.project = project_id
    request.instance_resource = instance

    # Wait for the create operation to complete.
    print(f"Creating the {instance_name} instance in {zone}...")

    operation = instance_client.insert(request=request)

    wait_for_extended_operation(operation, "instance creation")

    print(f"Instance {instance_name} created.")
    return instance_client.get(project=project_id, zone=zone, instance=instance_name)


def create_custom_instance(
    project_id: str,
    zone: str,
    instance_name: str,
    cpu_series: CustomMachineType.CPUSeries,
    core_count: int,
    memory: int,
) -> compute_v1.Instance:
    """
    Create a new VM instance with a custom machine type.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        zone: name of the zone to create the instance in. For example: "us-west3-b"
        instance_name: name of the new virtual machine (VM) instance.
        cpu_series: the type of CPU you want to use. Select one value from the CustomMachineType.CPUSeries enum.
            For example: CustomMachineType.CPUSeries.N2
        core_count: number of CPU cores you want to use.
        memory: the amount of memory for the VM instance, in megabytes.

    Return:
        Instance object.
    """
    assert cpu_series in (
        CustomMachineType.CPUSeries.E2,
        CustomMachineType.CPUSeries.N1,
        CustomMachineType.CPUSeries.N2,
        CustomMachineType.CPUSeries.N2D,
    )
    custom_type = CustomMachineType(zone, cpu_series, memory, core_count)

    newest_debian = get_image_from_family(project="debian-cloud", family="debian-10")
    disk_type = f"zones/{zone}/diskTypes/pd-standard"
    disks = [disk_from_image(disk_type, 10, True, newest_debian.self_link)]

    return create_instance(project_id, zone, instance_name, disks, str(custom_type))


# [END compute_custom_machine_type_create_with_helper]



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import uuid

import google.auth
from google.cloud import compute_v1
import pytest

from ..disks.create_from_image import create_disk_from_image
from ..disks.delete import delete_disk
from ..images.create import create_image_from_disk
from ..images.create_from_image import create_image_from_image
from ..images.create_from_snapshot import create_image_from_snapshot
from ..images.delete import delete_image
from ..images.get import get_image
from ..images.get import get_image_from_family
from ..images.list import list_images
from ..images.set_deprecation_status import set_deprecation_status
from ..snapshots.create import create_snapshot
from ..snapshots.delete import delete_snapshot

PROJECT = google.auth.default()[1]
ZONE = 'europe-west2-c'


@pytest.fixture
def test_disk():
    """
    Get the newest version of debian 11 and make a disk from it.
    """
    new_debian = get_image_from_family('debian-cloud', 'debian-11')
    test_disk_name = "test-disk-" + uuid.uuid4().hex[:10]
    disk = create_disk_from_image(PROJECT, ZONE, test_disk_name,
                                  f"zones/{ZONE}/diskTypes/pd-standard",
                                  20, new_debian.self_link)
    yield disk
    delete_disk(PROJECT, ZONE, test_disk_name)


@pytest.fixture
def test_snapshot(test_disk):
    """
    Make a snapshot that will be deleted when tests are done.
    """
    test_snap_name = "test-snap-" + uuid.uuid4().hex[:10]
    snap = create_snapshot(PROJECT, test_disk.name, test_snap_name, zone=test_disk.zone.rsplit('/')[-1])
    yield snap
    delete_snapshot(PROJECT, snap.name)


@pytest.fixture
def autodelete_image_name():
    """
    Provide a name for an image that will be deleted after the test is done.
    """
    test_img_name = "test-img-" + uuid.uuid4().hex[:10]
    yield test_img_name

    delete_image(PROJECT, test_img_name)


@pytest.fixture()
def autodelete_image(autodelete_image_name):
    """
    An image that will be deleted after the test is done.
    """
    src_img = get_image_from_family('debian-cloud', 'debian-11')
    new_image = create_image_from_image(PROJECT, src_img.name, autodelete_image_name, 'debian-cloud',
                                        storage_location='eu')
    yield new_image


def test_list_images():
    images = list_images("debian-cloud")
    for img in images:
        assert img.kind == "compute#image"
        assert img.self_link.startswith(
            "https://www.googleapis.com/compute/v1/projects/debian-cloud/global/images/"
        )


def test_get_image():
    images = list_images("debian-cloud")
    image = next(iter(images))

    image2 = get_image("debian-cloud", image.name)

    assert image.name == image2.name


def test_create_delete_image(test_disk):
    test_image_name = "test-image-" + uuid.uuid4().hex[:10]
    new_image = create_image_from_disk(PROJECT, ZONE, test_disk.name, test_image_name)
    try:
        assert new_image.name == test_image_name
        assert new_image.disk_size_gb == 20
        assert isinstance(new_image, compute_v1.Image)
    finally:
        delete_image(PROJECT, test_image_name)

        for image in list_images(PROJECT):
            if image.name == test_image_name:
                pytest.fail(f"Image {test_image_name} should have been deleted.")


def test_image_from_image(autodelete_image_name):
    src_img = get_image_from_family('ubuntu-os-cloud', 'ubuntu-2204-lts')
    new_image = create_image_from_image(PROJECT, src_img.name, autodelete_image_name, 'ubuntu-os-cloud',
                                        guest_os_features=[compute_v1.GuestOsFeature.Type.MULTI_IP_SUBNET.name],
                                        storage_location='eu')

    assert new_image.storage_locations == ['eu']
    assert new_image.disk_size_gb == src_img.disk_size_gb
    assert new_image.name == autodelete_image_name
    assert any(feature.type_ == compute_v1.GuestOsFeature.Type.MULTI_IP_SUBNET.name for feature in new_image.guest_os_features)


def test_image_from_snapshot(test_snapshot, autodelete_image_name):
    img = create_image_from_snapshot(PROJECT, test_snapshot.name, autodelete_image_name,
                                     guest_os_features=[compute_v1.GuestOsFeature.Type.MULTI_IP_SUBNET.name],
                                     storage_location='us-central1')
    assert img.storage_locations == ['us-central1']
    assert img.name == autodelete_image_name
    assert any(
        feature.type_ == compute_v1.GuestOsFeature.Type.MULTI_IP_SUBNET.name for feature in img.guest_os_features)


def test_status_change(autodelete_image):
    set_deprecation_status(PROJECT, autodelete_image.name, compute_v1.DeprecationStatus.State.DEPRECATED)
    img = get_image(PROJECT, autodelete_image.name)
    assert img.name == autodelete_image.name
    assert img.deprecated.state == compute_v1.DeprecationStatus.State.DEPRECATED.name



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# This is an ingredient file. It is not meant to be run directly. Check the samples/snippets 
# folder for complete code samples that are ready to be used.
# Disabling flake8 for the ingredients file, as it would fail F821 - undefined name check.
# flake8: noqa
from google.cloud import compute_v1


# <INGREDIENT start_instance_with_encryption_key>
def start_instance_with_encryption_key(
    project_id: str, zone: str, instance_name: str, key: bytes
) -> None:
    """
    Starts a stopped Google Compute Engine instance (with encrypted disks).
    Args:
        project_id: project ID or project number of the Cloud project your instance belongs to.
        zone: name of the zone your instance belongs to.
        instance_name: name of the instance your want to start.
        key: bytes object representing a raw base64 encoded key to your machines boot disk.
            For more information about disk encryption see:
            https://cloud.google.com/compute/docs/disks/customer-supplied-encryption#specifications
    """
    instance_client = compute_v1.InstancesClient()

    instance_data = instance_client.get(
        project=project_id, zone=zone, instance=instance_name
    )

    # Prepare the information about disk encryption
    disk_data = compute_v1.CustomerEncryptionKeyProtectedDisk()
    disk_data.source = instance_data.disks[0].source
    disk_data.disk_encryption_key = compute_v1.CustomerEncryptionKey()
    # Use raw_key to send over the key to unlock the disk
    # To use a key stored in KMS, you need to provide `kms_key_name` and `kms_key_service_account`
    disk_data.disk_encryption_key.raw_key = key
    enc_data = compute_v1.InstancesStartWithEncryptionKeyRequest()
    enc_data.disks = [disk_data]

    operation = instance_client.start_with_encryption_key(
        project=project_id,
        zone=zone,
        instance=instance_name,
        instances_start_with_encryption_key_request_resource=enc_data,
    )

    wait_for_extended_operation(operation, "instance start (with encrypted disk)")
# </INGREDIENT>



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# flake8: noqa


# This file is automatically generated. Please do not modify it directly.
# Find the relevant recipe file in the samples/recipes or samples/ingredients
# directory and apply your changes there.


from __future__ import annotations

from collections.abc import Iterable

from google.cloud import compute_v1


def create_firewall_rule(
    project_id: str, firewall_rule_name: str, network: str = "global/networks/default"
) -> compute_v1.Firewall:
    """
    Creates a simple firewall rule allowing for incoming HTTP and HTTPS access from the entire Internet.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        firewall_rule_name: name of the rule that is created.
        network: name of the network the rule will be applied to. Available name formats:
            * https://www.googleapis.com/compute/v1/projects/{project_id}/global/networks/{network}
            * projects/{project_id}/global/networks/{network}
            * global/networks/{network}

    Returns:
        A Firewall object.
    """
    firewall_rule = compute_v1.Firewall()
    firewall_rule.name = firewall_rule_name
    firewall_rule.direction = "INGRESS"

    allowed_ports = compute_v1.Allowed()
    allowed_ports.I_p_protocol = "tcp"
    allowed_ports.ports = ["80", "443"]

    firewall_rule.allowed = [allowed_ports]
    firewall_rule.source_ranges = ["0.0.0.0/0"]
    firewall_rule.network = network
    firewall_rule.description = "Allowing TCP traffic on port 80 and 443 from Internet."

    firewall_rule.target_tags = ["web"]

    # Note that the default value of priority for the firewall API is 1000.
    # If you check the value of `firewall_rule.priority` at this point it
    # will be equal to 0, however it is not treated as "set" by the library and thus
    # the default will be applied to the new rule. If you want to create a rule that
    # has priority == 0, you need to explicitly set it so:
    # TODO: Uncomment to set the priority to 0
    # firewall_rule.priority = 0

    firewall_client = compute_v1.FirewallsClient()
    operation = firewall_client.insert(
        project=project_id, firewall_resource=firewall_rule
    )

    wait_for_extended_operation(operation, "firewall rule creation")

    return firewall_client.get(project=project_id, firewall=firewall_rule_name)


def delete_firewall_rule(project_id: str, firewall_rule_name: str) -> None:
    """
    Deletes a firewall rule from the project.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        firewall_rule_name: name of the firewall rule you want to delete.
    """
    firewall_client = compute_v1.FirewallsClient()
    operation = firewall_client.delete(project=project_id, firewall=firewall_rule_name)

    wait_for_extended_operation(operation, "firewall rule deletion")


def get_firewall_rule(project_id: str, firewall_rule_name: str) -> compute_v1.Firewall:
    """
    Retrieve a Firewall from a project.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        firewall_rule_name: name of the firewall rule you want to retrieve.

    Returns:
        A Firewall object.
    """
    firewall_client = compute_v1.FirewallsClient()
    return firewall_client.get(project=project_id, firewall=firewall_rule_name)


def list_firewall_rules(project_id: str) -> Iterable[compute_v1.Firewall]:
    """
    Return a list of all the firewall rules in specified project. Also prints the
    list of firewall names and their descriptions.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.

    Returns:
        A flat list of all firewall rules defined for given project.
    """
    firewall_client = compute_v1.FirewallsClient()
    firewalls_list = firewall_client.list(project=project_id)

    for firewall in firewalls_list:
        print(f" - {firewall.name}: {firewall.description}")

    return firewalls_list


def patch_firewall_priority(
    project_id: str, firewall_rule_name: str, priority: int
) -> None:
    """
    Modifies the priority of a given firewall rule.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        firewall_rule_name: name of the rule you want to modify.
        priority: the new priority to be set for the rule.
    """
    firewall_rule = compute_v1.Firewall()
    firewall_rule.priority = priority

    # The patch operation doesn't require the full definition of a Firewall object. It will only update
    # the values that were set in it, in this case it will only change the priority.
    firewall_client = compute_v1.FirewallsClient()
    operation = firewall_client.patch(
        project=project_id, firewall=firewall_rule_name, firewall_resource=firewall_rule
    )

    wait_for_extended_operation(operation, "firewall rule patching")


if __name__ == "__main__":
    import google.auth
    import google.auth.exceptions

    try:
        default_project_id = google.auth.default()[1]
        print(f"Using project {default_project_id}.")
    except google.auth.exceptions.DefaultCredentialsError:
        print(
            "Please use `gcloud auth application-default login` "
            "or set GOOGLE_APPLICATION_CREDENTIALS to use this script."
        )
    else:
        import uuid

        rule_name = "firewall-sample-" + uuid.uuid4().hex[:10]
        print(f"Creating firewall rule {rule_name}...")
        # The rule will be created with default priority of 1000.
        create_firewall_rule(default_project_id, rule_name)
        try:
            print("Rule created:")
            print(get_firewall_rule(default_project_id, rule_name))
            print("Updating rule priority to 10...")
            patch_firewall_priority(default_project_id, rule_name, 10)
            print("Rule updated: ")
            print(get_firewall_rule(default_project_id, rule_name))
            print(f"Deleting rule {rule_name}...")
        finally:
            delete_firewall_rule(default_project_id, rule_name)
        print("Done.")



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# flake8: noqa


# This file is automatically generated. Please do not modify it directly.
# Find the relevant recipe file in the samples/recipes or samples/ingredients
# directory and apply your changes there.


# [START compute_create_route_windows_activation]
from __future__ import annotations

import sys
from typing import Any

from google.api_core.extended_operation import ExtendedOperation
from google.cloud import compute_v1


def wait_for_extended_operation(
    operation: ExtendedOperation, verbose_name: str = "operation", timeout: int = 300
) -> Any:
    """
    Waits for the extended (long-running) operation to complete.

    If the operation is successful, it will return its result.
    If the operation ends with an error, an exception will be raised.
    If there were any warnings during the execution of the operation
    they will be printed to sys.stderr.

    Args:
        operation: a long-running operation you want to wait on.
        verbose_name: (optional) a more verbose name of the operation,
            used only during error and warning reporting.
        timeout: how long (in seconds) to wait for operation to finish.
            If None, wait indefinitely.

    Returns:
        Whatever the operation.result() returns.

    Raises:
        This method will raise the exception received from `operation.exception()`
        or RuntimeError if there is no exception set, but there is an `error_code`
        set for the `operation`.

        In case of an operation taking longer than `timeout` seconds to complete,
        a `concurrent.futures.TimeoutError` will be raised.
    """
    result = operation.result(timeout=timeout)

    if operation.error_code:
        print(
            f"Error during {verbose_name}: [Code: {operation.error_code}]: {operation.error_message}",
            file=sys.stderr,
            flush=True,
        )
        print(f"Operation ID: {operation.name}", file=sys.stderr, flush=True)
        raise operation.exception() or RuntimeError(operation.error_message)

    if operation.warnings:
        print(f"Warnings during {verbose_name}:\n", file=sys.stderr, flush=True)
        for warning in operation.warnings:
            print(f" - {warning.code}: {warning.message}", file=sys.stderr, flush=True)

    return result


def create_route(
    project_id: str,
    network: str,
    route_name: str,
    destination_range: str,
    *,
    next_hop_gateway: str | None = None,
    next_hop_ip: str | None = None,
    next_hop_instance: str | None = None,
    next_hop_vpn_tunnel: str | None = None,
    next_hop_ilb: str | None = None,
) -> compute_v1.Route:
    """
    Create a new route in selected network by providing a destination and next hop name.

    Note: The set of {next_hop_gateway, next_hop_ip, next_hop_instance, next_hop_vpn_tunnel,
        next_hop_ilb} is exclusive, you and only specify one of those parameters.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        network: name of the network the route will be created in. Available name formats:
            * https://www.googleapis.com/compute/v1/projects/{project_id}/global/networks/{network}
            * projects/{project_id}/global/networks/{network}
            * global/networks/{network}
        route_name: name of the new route.
        destination_range: range of destination IPs this route should be applied to. E.g. 10.0.0.0/16.
        next_hop_gateway: name of the gateway the traffic should be directed to.
        next_hop_ip: IP address the traffic should be directed to.
        next_hop_instance: name of the instance the traffic should be directed to. Name format:
            "projects/{project}/zones/{zone}/instances/{instance_name}"
        next_hop_vpn_tunnel: name of the VPN tunnel the traffic should be directed to. Name format:
            "projects/{project}/regions/{region}/vpnTunnels/{vpn_tunnel_name}"
        next_hop_ilb: name of a forwarding rule of the Internal Load Balancer the traffic
            should be directed to. Name format:
            "projects/{project}/regions/{region}/forwardingRules/{forwarding_rule_region}"

    Returns:
        A new compute_v1.Route object.
    """
    excl_args = {
        next_hop_instance,
        next_hop_ilb,
        next_hop_vpn_tunnel,
        next_hop_gateway,
        next_hop_ip,
    }
    args_set = sum(1 if arg is not None else 0 for arg in excl_args)

    if args_set != 1:
        raise RuntimeError("You must specify exactly one next_hop_* parameter.")

    route = compute_v1.Route()
    route.name = route_name
    route.network = network
    route.dest_range = destination_range

    if next_hop_gateway:
        route.next_hop_gateway = next_hop_gateway
    elif next_hop_ip:
        route.next_hop_ip = next_hop_ip
    elif next_hop_instance:
        route.next_hop_instance = next_hop_instance
    elif next_hop_vpn_tunnel:
        route.next_hop_vpn_tunnel = next_hop_vpn_tunnel
    elif next_hop_ilb:
        route.next_hop_ilb = next_hop_ilb

    route_client = compute_v1.RoutesClient()
    operation = route_client.insert(project=project_id, route_resource=route)

    wait_for_extended_operation(operation, "route creation")

    return route_client.get(project=project_id, route=route_name)


def create_route_to_windows_activation_host(
    project_id: str, network: str, route_name: str
) -> compute_v1.Route:
    """
    If you have Windows instances without external IP addresses,
    you must also enable Private Google Access so that instances
    with only internal IP addresses can send traffic to the external
    IP address for kms.windows.googlecloud.com.
    More infromation: https://cloud.google.com/vpc/docs/configure-private-google-access#enabling

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        network: name of the network the route will be created in. Available name formats:
            * https://www.googleapis.com/compute/v1/projects/{project_id}/global/networks/{network}
            * projects/{project_id}/global/networks/{network}
            * global/networks/{network}
        route_name: name of the new route.

    Returns:
        A new compute_v1.Route object.
    """
    return create_route(
        project_id=project_id,
        network=network,
        route_name=route_name,
        destination_range="35.190.247.13/32",
        next_hop_gateway=f"projects/{project_id}/global/gateways/default-internet-gateway",
    )


# [END compute_create_route_windows_activation]



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import uuid

import google.auth
import pytest

from ..instances.delete import delete_instance
from ..instances.delete_protection.create import create_protected_instance
from ..instances.delete_protection.get import get_delete_protection
from ..instances.delete_protection.set import set_delete_protection

PROJECT = google.auth.default()[1]
INSTANCE_ZONE = "europe-west2-a"


@pytest.fixture
def autodelete_instance_name():
    instance_name = "test-instance-" + uuid.uuid4().hex[:10]

    yield instance_name

    if get_delete_protection(PROJECT, INSTANCE_ZONE, instance_name):
        set_delete_protection(PROJECT, INSTANCE_ZONE, instance_name, False)

    delete_instance(PROJECT, INSTANCE_ZONE, instance_name)


def test_delete_protection(autodelete_instance_name):
    instance = create_protected_instance(
        PROJECT, INSTANCE_ZONE, autodelete_instance_name
    )
    assert instance.name == autodelete_instance_name

    assert (
        get_delete_protection(PROJECT, INSTANCE_ZONE, autodelete_instance_name) is True
    )

    set_delete_protection(PROJECT, INSTANCE_ZONE, autodelete_instance_name, False)

    assert (
        get_delete_protection(PROJECT, INSTANCE_ZONE, autodelete_instance_name) is False
    )



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# This is an ingredient file. It is not meant to be run directly. Check the samples/snippets 
# folder for complete code samples that are ready to be used.
# Disabling flake8 for the ingredients file, as it would fail F821 - undefined name check.
# flake8: noqa

from google.cloud import compute_v1


# <INGREDIENT create_template_from_instance>
def create_template_from_instance(
    project_id: str, instance: str, template_name: str
) -> compute_v1.InstanceTemplate:
    """
    Create a new instance template based on an existing instance.
    This new template specifies a different boot disk.

    Args:
        project_id: project ID or project number of the Cloud project you use.
        instance: the instance to base the new template on. This value uses
            the following format: "projects/{project}/zones/{zone}/instances/{instance_name}"
        template_name: name of the new template to create.

    Returns:
        InstanceTemplate object that represents the new instance template.
    """
    disk = compute_v1.DiskInstantiationConfig()
    # Device name must match the name of a disk attached to the instance you are
    # basing your template on.
    disk.device_name = "disk-1"
    # Replace the original boot disk image used in your instance with a Rocky Linux image.
    disk.instantiate_from = "CUSTOM_IMAGE"
    disk.custom_image = "projects/rocky-linux-cloud/global/images/family/rocky-linux-8"
    # Override the auto_delete setting.
    disk.auto_delete = True

    template = compute_v1.InstanceTemplate()
    template.name = template_name
    template.source_instance = instance
    template.source_instance_params = compute_v1.SourceInstanceParams()
    template.source_instance_params.disk_configs = [disk]

    template_client = compute_v1.InstanceTemplatesClient()
    operation = template_client.insert(
        project=project_id, instance_template_resource=template
    )

    wait_for_extended_operation(operation, "instance template creation")

    return template_client.get(project=project_id, instance_template=template_name)
# </INGREDIENT>



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from copy import deepcopy
import uuid

import google.api_core.exceptions
import google.auth
from google.cloud import compute_v1
import pytest


from ..disks.delete import delete_disk
from ..disks.list import list_disks
from ..instances.create_start_instance.create_windows_instance import \
    get_image_from_family

PROJECT = google.auth.default()[1]
ZONE = "europe-west1-c"


def test_request_id():
    disk = compute_v1.Disk()
    disk.size_gb = 20
    disk.name = "test-disk-" + uuid.uuid4().hex[:10]
    disk.zone = ZONE
    disk.type_ = f"zones/{ZONE}/diskTypes/pd-standard"
    disk.source_image = get_image_from_family("debian-cloud", "debian-11").self_link

    disk2 = deepcopy(disk)
    disk2.name = "test-disk-" + uuid.uuid4().hex[:10]

    request = compute_v1.InsertDiskRequest()
    request.request_id = str(uuid.uuid4())
    request.project = PROJECT
    request.zone = ZONE
    request.disk_resource = disk

    # Creating a different request, but with the same request_id
    # This should not be executed, because the previous request
    # has the same ID.
    request2 = deepcopy(request)
    request2.disk_resource = disk2

    disk_client = compute_v1.DisksClient()
    try:
        operation = disk_client.insert(request)
        operation2 = disk_client.insert(request2)
        operation.result()
        operation2.result()
    except Exception as err:
        pytest.fail(f"There was an error: {err}")
        raise err
    else:
        disks = list_disks(PROJECT, ZONE)
        assert any(i_disk.name == disk.name for i_disk in disks)
        assert all(i_disk.name != disk2.name for i_disk in disks)
    finally:
        delete_disk(PROJECT, ZONE, disk.name)
        try:
            delete_disk(PROJECT, ZONE, disk2.name)
        except google.api_core.exceptions.NotFound:
            pass


def test_request_id_op_id():
    disk = compute_v1.Disk()
    disk.size_gb = 20
    disk.name = "test-disk-" + uuid.uuid4().hex[:10]
    disk.zone = ZONE
    disk.type_ = f"zones/{ZONE}/diskTypes/pd-standard"
    disk.source_image = get_image_from_family("debian-cloud", "debian-11").self_link

    request = compute_v1.InsertDiskRequest()
    request.request_id = str(uuid.uuid4())
    request.project = PROJECT
    request.zone = ZONE
    request.disk_resource = disk

    disk_client = compute_v1.DisksClient()

    try:
        op1 = disk_client.insert(request)
        op2 = disk_client.insert(request)
        op1.result()
        assert op1.name == op2.name
    finally:
        delete_disk(PROJECT, ZONE, disk.name)



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# flake8: noqa


# This file is automatically generated. Please do not modify it directly.
# Find the relevant recipe file in the samples/recipes or samples/ingredients
# directory and apply your changes there.


# [START compute_instances_bulk_insert]
from __future__ import annotations

from collections.abc import Iterable
import sys
from typing import Any
import uuid

from google.api_core.extended_operation import ExtendedOperation
from google.cloud import compute_v1


def wait_for_extended_operation(
    operation: ExtendedOperation, verbose_name: str = "operation", timeout: int = 300
) -> Any:
    """
    Waits for the extended (long-running) operation to complete.

    If the operation is successful, it will return its result.
    If the operation ends with an error, an exception will be raised.
    If there were any warnings during the execution of the operation
    they will be printed to sys.stderr.

    Args:
        operation: a long-running operation you want to wait on.
        verbose_name: (optional) a more verbose name of the operation,
            used only during error and warning reporting.
        timeout: how long (in seconds) to wait for operation to finish.
            If None, wait indefinitely.

    Returns:
        Whatever the operation.result() returns.

    Raises:
        This method will raise the exception received from `operation.exception()`
        or RuntimeError if there is no exception set, but there is an `error_code`
        set for the `operation`.

        In case of an operation taking longer than `timeout` seconds to complete,
        a `concurrent.futures.TimeoutError` will be raised.
    """
    result = operation.result(timeout=timeout)

    if operation.error_code:
        print(
            f"Error during {verbose_name}: [Code: {operation.error_code}]: {operation.error_message}",
            file=sys.stderr,
            flush=True,
        )
        print(f"Operation ID: {operation.name}", file=sys.stderr, flush=True)
        raise operation.exception() or RuntimeError(operation.error_message)

    if operation.warnings:
        print(f"Warnings during {verbose_name}:\n", file=sys.stderr, flush=True)
        for warning in operation.warnings:
            print(f" - {warning.code}: {warning.message}", file=sys.stderr, flush=True)

    return result


def get_instance_template(
    project_id: str, template_name: str
) -> compute_v1.InstanceTemplate:
    """
    Retrieve an instance template, which you can use to create virtual machine
    (VM) instances and managed instance groups (MIGs).

    Args:
        project_id: project ID or project number of the Cloud project you use.
        template_name: name of the template to retrieve.

    Returns:
        InstanceTemplate object that represents the retrieved template.
    """
    template_client = compute_v1.InstanceTemplatesClient()
    return template_client.get(project=project_id, instance_template=template_name)


def bulk_insert_instance(
    project_id: str,
    zone: str,
    template: compute_v1.InstanceTemplate,
    count: int,
    name_pattern: str,
    min_count: int | None = None,
    labels: dict | None = None,
) -> Iterable[compute_v1.Instance]:
    """
    Create multiple VMs based on an Instance Template. The newly created instances will
    be returned as a list and will share a label with key `bulk_batch` and a random
    value.

    If the bulk insert operation fails and the requested number of instances can't be created,
    and more than min_count instances are created, then those instances can be found using
    the `bulk_batch` label with value attached to the raised exception in bulk_batch_id
    attribute. So, you can use the following filter: f"label.bulk_batch={err.bulk_batch_id}"
    when listing instances in a zone to get the instances that were successfully created.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        zone: name of the zone to create the instance in. For example: "us-west3-b"
        template: an Instance Template to be used for creation of the new VMs.
        name_pattern: The string pattern used for the names of the VMs. The pattern
            must contain one continuous sequence of placeholder hash characters (#)
            with each character corresponding to one digit of the generated instance
            name. Example: a name_pattern of inst-#### generates instance names such
            as inst-0001 and inst-0002. If existing instances in the same project and
            zone have names that match the name pattern then the generated instance
            numbers start after the biggest existing number. For example, if there
            exists an instance with name inst-0050, then instance names generated
            using the pattern inst-#### begin with inst-0051. The name pattern
            placeholder #...# can contain up to 18 characters.
        count: The maximum number of instances to create.
        min_count (optional): The minimum number of instances to create. If no min_count is
            specified then count is used as the default value. If min_count instances
            cannot be created, then no instances will be created and instances already
            created will be deleted.
        labels (optional): A dictionary with labels to be added to the new VMs.
    """
    bulk_insert_resource = compute_v1.BulkInsertInstanceResource()
    bulk_insert_resource.source_instance_template = template.self_link
    bulk_insert_resource.count = count
    bulk_insert_resource.min_count = min_count or count
    bulk_insert_resource.name_pattern = name_pattern

    if not labels:
        labels = {}

    labels["bulk_batch"] = uuid.uuid4().hex
    instance_prop = compute_v1.InstanceProperties()
    instance_prop.labels = labels
    bulk_insert_resource.instance_properties = instance_prop

    bulk_insert_request = compute_v1.BulkInsertInstanceRequest()
    bulk_insert_request.bulk_insert_instance_resource_resource = bulk_insert_resource
    bulk_insert_request.project = project_id
    bulk_insert_request.zone = zone

    client = compute_v1.InstancesClient()
    operation = client.bulk_insert(bulk_insert_request)

    try:
        wait_for_extended_operation(operation, "bulk instance creation")
    except Exception as err:
        err.bulk_batch_id = labels["bulk_batch"]
        raise err

    list_req = compute_v1.ListInstancesRequest()
    list_req.project = project_id
    list_req.zone = zone
    list_req.filter = " AND ".join(
        f"labels.{key}:{value}" for key, value in labels.items()
    )
    return client.list(list_req)


def create_five_instances(
    project_id: str, zone: str, template_name: str, name_pattern: str
):
    """
    Create five instances of an instance template.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        zone: name of the zone to create the instance in. For example: "us-west3-b"
        template_name: name of the template that will be used to create new VMs.
        name_pattern: The string pattern used for the names of the VMs.
    """
    template = get_instance_template(project_id, template_name)
    instances = bulk_insert_instance(project_id, zone, template, 5, name_pattern)
    return instances


# [END compute_instances_bulk_insert]



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# flake8: noqa


# This file is automatically generated. Please do not modify it directly.
# Find the relevant recipe file in the samples/recipes or samples/ingredients
# directory and apply your changes there.


# [START compute_images_get_from_family]
# [START compute_images_get]
from google.cloud import compute_v1

# [END compute_images_get]


def get_image_from_family(project: str, family: str) -> compute_v1.Image:
    """
    Retrieve the newest image that is part of a given family in a project.

    Args:
        project: project ID or project number of the Cloud project you want to get image from.
        family: name of the image family you want to get image from.

    Returns:
        An Image object.
    """
    image_client = compute_v1.ImagesClient()
    # List of public operating system (OS) images: https://cloud.google.com/compute/docs/images/os-details
    newest_image = image_client.get_from_family(project=project, family=family)
    return newest_image


# [END compute_images_get_from_family]


# [START compute_images_get]
def get_image(project_id: str, image_name: str) -> compute_v1.Image:
    """
    Retrieve detailed information about a single image from a project.

    Args:
        project_id: project ID or project number of the Cloud project you want to list images from.
        image_name: name of the image you want to get details of.

    Returns:
        An instance of compute_v1.Image object with information about specified image.
    """
    image_client = compute_v1.ImagesClient()
    return image_client.get(project=project_id, image=image_name)


# [END compute_images_get]



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

import google.auth
import pytest

# Turning off F401 check because flake8 doesn't recognize using
# PyTest fixture as parameter as usage.
from .test_instance_start_stop import compute_instance  # noqa: F401

from ..instance_templates.create import create_template
from ..instance_templates.create_from_instance import \
    create_template_from_instance
from ..instance_templates.create_with_subnet import create_template_with_subnet
from ..instance_templates.delete import delete_instance_template
from ..instance_templates.list import list_instance_templates

PROJECT = google.auth.default()[1]

INSTANCE_ZONE = "europe-west2-b"


@pytest.fixture
def deletable_template_name():
    template_name = "i" + uuid.uuid4().hex[:10]
    yield template_name
    delete_instance_template(PROJECT, template_name)


@pytest.fixture
def template_to_be_deleted():
    template_name = "i" + uuid.uuid4().hex[:10]
    template = create_template(PROJECT, template_name)
    yield template


def test_create_template_and_list(deletable_template_name):

    template = create_template(PROJECT, deletable_template_name)

    assert template.name == deletable_template_name
    assert any(
        template.name == deletable_template_name
        for template in list_instance_templates(PROJECT)
    )
    assert template.properties.disks[0].initialize_params.disk_size_gb == 250
    assert "debian-11" in template.properties.disks[0].initialize_params.source_image
    assert template.properties.network_interfaces[0].name == "global/networks/default"
    assert template.properties.machine_type == "e2-standard-4"


def test_create_from_instance(compute_instance, deletable_template_name):  # noqa: F811

    template = create_template_from_instance(
        PROJECT, compute_instance.self_link, deletable_template_name
    )

    assert template.name == deletable_template_name
    assert template.properties.machine_type in compute_instance.machine_type
    assert (
        template.properties.disks[0].disk_size_gb
        == compute_instance.disks[0].disk_size_gb
    )
    assert (
        template.properties.disks[0].initialize_params.source_image
        == "projects/rocky-linux-cloud/global/images/family/rocky-linux-8"
    )


def test_create_template_with_subnet(deletable_template_name):
    template = create_template_with_subnet(
        PROJECT,
        "global/networks/default",
        "regions/asia-east1/subnetworks/default",
        deletable_template_name,
    )

    assert template.name == deletable_template_name
    assert (
        "global/networks/default" in template.properties.network_interfaces[0].network
    )
    assert (
        "regions/asia-east1/subnetworks/default"
        in template.properties.network_interfaces[0].subnetwork
    )


def test_delete_template(template_to_be_deleted):
    delete_instance_template(PROJECT, template_to_be_deleted.name)

    assert all(
        template.name != template_to_be_deleted.name
        for template in list_instance_templates(PROJECT)
    )



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# This is an ingredient file. It is not meant to be run directly. Check the samples/snippets
# folder for complete code samples that are ready to be used.
# Disabling flake8 for the ingredients file, as it would fail F821 - undefined name check.
# flake8: noqa

from __future__ import annotations

from collections.abc import Iterable
from google.cloud import compute_v1


# <INGREDIENT list_instance_templates>
def list_instance_templates(project_id: str) -> Iterable[compute_v1.InstanceTemplate]:
    """
    Get a list of InstanceTemplate objects available in a project.

    Args:
        project_id: project ID or project number of the Cloud project you use.

    Returns:
        Iterable list of InstanceTemplate objects.
    """
    template_client = compute_v1.InstanceTemplatesClient()
    return template_client.list(project=project_id)
# </INGREDIENT>



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# This is an ingredient file. It is not meant to be run directly. Check the samples/snippets
# folder for complete code samples that are ready to be used.
# Disabling flake8 for the ingredients file, as it would fail F821 - undefined name check.
# flake8: noqa
from __future__ import annotations


from google.cloud import compute_v1


# <INGREDIENT create_snapshot>
def create_snapshot(project_id: str, disk_name: str, snapshot_name: str, *,
                    zone: str | None = None, region: str | None = None,
                    location: str | None = None, disk_project_id: str | None = None) -> compute_v1.Snapshot:
    """
    Create a snapshot of a disk.

    You need to pass `zone` or `region` parameter relevant to the disk you want to
    snapshot, but not both. Pass `zone` parameter for zonal disks and `region` for
    regional disks.

    Args:
        project_id: project ID or project number of the Cloud project you want
            to use to store the snapshot.
        disk_name: name of the disk you want to snapshot.
        snapshot_name: name of the snapshot to be created.
        zone: name of the zone in which is the disk you want to snapshot (for zonal disks).
        region: name of the region in which is the disk you want to snapshot (for regional disks).
        location: The Cloud Storage multi-region or the Cloud Storage region where you
            want to store your snapshot.
            You can specify only one storage location. Available locations:
            https://cloud.google.com/storage/docs/locations#available-locations
        disk_project_id: project ID or project number of the Cloud project that
            hosts the disk you want to snapshot. If not provided, will look for
            the disk in the `project_id` project.

    Returns:
        The new snapshot instance.
    """
    if zone is None and region is None:
        raise RuntimeError("You need to specify `zone` or `region` for this function to work.")
    if zone is not None and region is not None:
        raise RuntimeError("You can't set both `zone` and `region` parameters.")

    if disk_project_id is None:
        disk_project_id = project_id

    if zone is not None:
        disk_client = compute_v1.DisksClient()
        disk = disk_client.get(project=disk_project_id, zone=zone, disk=disk_name)
    else:
        regio_disk_client = compute_v1.RegionDisksClient()
        disk = regio_disk_client.get(project=disk_project_id, region=region, disk=disk_name)

    snapshot = compute_v1.Snapshot()
    snapshot.source_disk = disk.self_link
    snapshot.name = snapshot_name
    if location:
        snapshot.storage_locations = [location]

    snapshot_client = compute_v1.SnapshotsClient()
    operation = snapshot_client.insert(project=project_id, snapshot_resource=snapshot)

    wait_for_extended_operation(operation, "snapshot creation")

    return snapshot_client.get(project=project_id, snapshot=snapshot_name)

# </INGREDIENT>



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# This is an ingredient file. It is not meant to be run directly. Check the samples/snippets 
# folder for complete code samples that are ready to be used.
# Disabling flake8 for the ingredients file, as it would fail F821 - undefined name check.
# flake8: noqa
from google.cloud import compute_v1

from google.cloud.compute_v1.services.zone_operations import pagers


# <INGREDIENT list_zone_operations>
def list_zone_operations(
    project_id: str, zone: str, filter: str = ""
) -> pagers.ListPager:
    """
    List all recent operations the happened in given zone in a project. Optionally filter those
    operations by providing a filter. More about using the filter can be found here:
    https://cloud.google.com/compute/docs/reference/rest/v1/zoneOperations/list
    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        zone: name of the zone you want to use. For example: "us-west3-b"
        filter: filter string to be used for this listing operation.
    Returns:
        List of preemption operations in given zone.
    """
    operation_client = compute_v1.ZoneOperationsClient()
    request = compute_v1.ListZoneOperationsRequest()
    request.project = project_id
    request.zone = zone
    request.filter = filter

    return operation_client.list(request)
# </INGREDIENT>



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# This is an ingredient file. It is not meant to be run directly. Check the samples/snippets 
# folder for complete code samples that are ready to be used.
# Disabling flake8 for the ingredients file, as it would fail F821 - undefined name check.
# flake8: noqa

from google.cloud import compute_v1


# <INGREDIENT create_instance_from_template>
def create_instance_from_template(
    project_id: str, zone: str, instance_name: str, instance_template_url: str
) -> compute_v1.Instance:
    """
    Creates a Compute Engine VM instance from an instance template.

    Args:
        project_id: ID or number of the project you want to use.
        zone: Name of the zone you want to check, for example: us-west3-b
        instance_name: Name of the new instance.
        instance_template_url: URL of the instance template used for creating the new instance.
            It can be a full or partial URL.
            Examples:
            - https://www.googleapis.com/compute/v1/projects/project/global/instanceTemplates/example-instance-template
            - projects/project/global/instanceTemplates/example-instance-template
            - global/instanceTemplates/example-instance-template

    Returns:
        Instance object.
    """
    instance_client = compute_v1.InstancesClient()

    instance_insert_request = compute_v1.InsertInstanceRequest()
    instance_insert_request.project = project_id
    instance_insert_request.zone = zone
    instance_insert_request.source_instance_template = instance_template_url
    instance_insert_request.instance_resource.name = instance_name

    operation = instance_client.insert(instance_insert_request)
    wait_for_extended_operation(operation, "instance creation")

    return instance_client.get(project=project_id, zone=zone, instance=instance_name)
# </INGREDIENT>



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import time
import uuid

from google.api_core.exceptions import NotFound
import google.auth
from google.cloud import compute_v1, kms_v1
import pytest

from ..disks.attach_disk import attach_disk
from ..disks.clone_encrypted_disk_managed_key import \
    create_disk_from_kms_encrypted_disk
from ..disks.create_empty_disk import create_empty_disk
from ..disks.create_from_image import create_disk_from_image
from ..disks.create_from_source import create_disk_from_disk
from ..disks.create_kms_encrypted_disk import create_kms_encrypted_disk
from ..disks.delete import delete_disk
from ..disks.list import list_disks
from ..disks.regional_create_from_source import create_regional_disk
from ..disks.regional_delete import delete_regional_disk
from ..disks.resize_disk import resize_disk
from ..images.get import get_image_from_family
from ..instances.create import create_instance, disk_from_image
from ..instances.delete import delete_instance
from ..instances.get import get_instance
from ..snapshots.create import create_snapshot
from ..snapshots.delete import delete_snapshot

PROJECT = google.auth.default()[1]
ZONE = 'europe-west2-c'
REGION = 'europe-west2'
KMS_KEYRING_NAME = 'compute-test-keyring'
KMS_KEY_NAME = 'compute-test-key'


@pytest.fixture()
def kms_key():
    client = kms_v1.KeyManagementServiceClient()
    location = f"projects/{PROJECT}/locations/global"
    keyring_link = f"projects/{PROJECT}/locations/global/keyRings/{KMS_KEYRING_NAME}"
    key_name = f"{keyring_link}/cryptoKeys/{KMS_KEY_NAME}"

    for ring in client.list_key_rings(parent=location):
        if ring.name == keyring_link:
            break
    else:
        client.create_key_ring(parent=location, key_ring_id=KMS_KEYRING_NAME)

    for key in client.list_crypto_keys(parent=keyring_link):
        if key.name == key_name:
            break
    else:
        key = kms_v1.CryptoKey()
        key.purpose = key.CryptoKeyPurpose.ENCRYPT_DECRYPT
        client.create_crypto_key(parent=keyring_link, crypto_key_id=KMS_KEY_NAME, crypto_key=key)

    yield client.get_crypto_key(name=key_name)


@pytest.fixture
def test_disk():
    """
    Get the newest version of debian 11 and make a disk from it.
    """
    new_debian = get_image_from_family('debian-cloud', 'debian-11')
    test_disk_name = "test-disk-" + uuid.uuid4().hex[:10]
    disk = create_disk_from_image(PROJECT, ZONE, test_disk_name,
                                  f"zones/{ZONE}/diskTypes/pd-standard",
                                  20, new_debian.self_link)
    yield disk
    delete_disk(PROJECT, ZONE, test_disk_name)


@pytest.fixture
def test_snapshot(test_disk):
    """
    Make a snapshot that will be deleted when tests are done.
    """
    test_snap_name = "test-snap-" + uuid.uuid4().hex[:10]
    snap = create_snapshot(PROJECT, test_disk.name, test_snap_name, zone=test_disk.zone.rsplit('/')[-1])
    yield snap
    delete_snapshot(PROJECT, snap.name)


@pytest.fixture()
def autodelete_disk_name():
    disk_name = "test-disk-" + uuid.uuid4().hex[:10]
    yield disk_name
    try:
        delete_disk(PROJECT, ZONE, disk_name)
    except NotFound:
        # The disk was already deleted
        pass


# To use the fixture 2 times in one test:
# https://stackoverflow.com/questions/36100624/pytest-use-same-fixture-twice-in-one-function
autodelete_disk_name2 = autodelete_disk_name


@pytest.fixture()
def autodelete_src_disk(autodelete_disk_name):
    disk_type = f"zones/{ZONE}/diskTypes/pd-standard"
    debian_image = get_image_from_family('debian-cloud', 'debian-11')
    disk = create_disk_from_image(PROJECT, ZONE, autodelete_disk_name, disk_type, 24, debian_image.self_link)
    yield disk


@pytest.fixture
def autodelete_instance_name():
    instance_name = "test-instance-" + uuid.uuid4().hex[:10]

    yield instance_name

    delete_instance(PROJECT, ZONE, instance_name)


@pytest.fixture
def autodelete_regional_blank_disk():
    disk_name = 'regional-disk-' + uuid.uuid4().hex[:10]
    replica_zones = [
        f"projects/{PROJECT}/zones/{REGION}-c",
        f"projects/{PROJECT}/zones/{REGION}-b",
    ]
    disk_type = f"regions/{REGION}/diskTypes/pd-balanced"

    disk = create_regional_disk(PROJECT, REGION, replica_zones, disk_name, disk_type, 11)

    yield disk

    try:
        # We wait for a while to let instances using this disk to be removed.
        time.sleep(60)
        delete_regional_disk(PROJECT, REGION, disk_name)
    except NotFound:
        # The disk was already deleted
        pass


@pytest.fixture
def autodelete_blank_disk():
    disk_name = 'regional-disk-' + uuid.uuid4().hex[:10]
    disk_type = f"zones/{ZONE}/diskTypes/pd-standard"

    disk = create_empty_disk(PROJECT, ZONE, disk_name, disk_type, 12)

    yield disk

    try:
        # We wait for a while to let instances using this disk to be removed.
        print('Waiting')
        time.sleep(60)
        print('Deleting')
        delete_disk(PROJECT, ZONE, disk_name)
    except NotFound:
        # The disk was already deleted
        pass


@pytest.fixture
def autodelete_compute_instance():
    instance_name = "test-instance-" + uuid.uuid4().hex[:10]
    newest_debian = get_image_from_family(project="ubuntu-os-cloud", family="ubuntu-2204-lts")
    disk_type = f"zones/{ZONE}/diskTypes/pd-standard"
    disks = [disk_from_image(disk_type, 100, True, newest_debian.self_link)]
    instance = create_instance(
        PROJECT, ZONE, instance_name, disks
    )
    yield instance

    delete_instance(PROJECT, ZONE, instance_name)


def test_disk_create_delete(autodelete_disk_name):
    disk_type = f"zones/{ZONE}/diskTypes/pd-standard"
    debian_image = get_image_from_family('debian-cloud', 'debian-11')

    disk = create_disk_from_image(PROJECT, ZONE, autodelete_disk_name, disk_type, 17, debian_image.self_link)
    assert disk.name == autodelete_disk_name
    assert disk.type_.endswith(disk_type)
    assert disk.size_gb == 17

    for i_disk in list_disks(PROJECT, ZONE):
        if i_disk.name == autodelete_disk_name:
            break
    else:
        pytest.fail("Couldn't find newly created disk on the disk list.")

    delete_disk(PROJECT, ZONE, autodelete_disk_name)

    for i_disk in list_disks(PROJECT, ZONE):
        if i_disk.name == autodelete_disk_name:
            pytest.fail("Found a disk that should be deleted on the disk list.")


def test_create_and_clone_encrypted_disk(autodelete_disk_name, kms_key, autodelete_disk_name2):
    # The service account service-{PROJECT_ID}@compute-system.iam.gserviceaccount.com needs to have the
    # cloudkms.cryptoKeyVersions.useToEncrypt permission to execute this test.
    # Best way is to give this account the cloudkms.cryptoKeyEncrypterDecrypter role.
    disk_type = f"zones/{ZONE}/diskTypes/pd-standard"
    debian_image = get_image_from_family('debian-cloud', 'debian-11')

    disk = create_kms_encrypted_disk(PROJECT, ZONE, autodelete_disk_name, disk_type, 25, kms_key.name,
                                     image_link=debian_image.self_link)
    assert disk.name == autodelete_disk_name
    assert disk.type_.endswith(disk_type)

    disk2 = create_disk_from_kms_encrypted_disk(PROJECT, ZONE, autodelete_disk_name2, disk_type,
                                                25, disk_link=disk.self_link, kms_key_name=kms_key.name)
    assert disk2.name == autodelete_disk_name2
    assert disk2.type_.endswith(disk_type)


def test_create_disk_from_disk(autodelete_src_disk, autodelete_disk_name2):
    disk_type = f"zones/{ZONE}/diskTypes/pd-standard"
    new_disk = create_disk_from_disk(PROJECT, ZONE, autodelete_disk_name2, disk_type, 24, autodelete_src_disk.self_link)

    assert new_disk.type_.endswith(disk_type)
    assert new_disk.name == autodelete_disk_name2


def test_create_and_delete_regional_disk(test_snapshot):
    disk_name = "test-rdisk-" + uuid.uuid4().hex[:10]
    disk_type = f"regions/{REGION}/diskTypes/pd-balanced"
    replica_zones = [
        f"projects/{PROJECT}/zones/{REGION}-a",
        f"projects/{PROJECT}/zones/{REGION}-b",
    ]

    try:
        regional_disk = create_regional_disk(PROJECT, REGION, replica_zones, disk_name,
                                             disk_type, 25, snapshot_link=test_snapshot.self_link)
        assert regional_disk.name == disk_name
        assert regional_disk.type_.endswith(disk_type)
    finally:
        delete_regional_disk(PROJECT, REGION, disk_name)


def test_disk_attachment(autodelete_blank_disk, autodelete_regional_blank_disk, autodelete_compute_instance):
    instance = get_instance(PROJECT, ZONE, autodelete_compute_instance.name)

    assert len(list(instance.disks)) == 1

    attach_disk(PROJECT, ZONE, instance.name, autodelete_blank_disk.self_link, 'READ_ONLY')
    attach_disk(PROJECT, ZONE, instance.name, autodelete_regional_blank_disk.self_link, 'READ_WRITE')

    instance = get_instance(PROJECT, ZONE, autodelete_compute_instance.name)

    assert len(list(instance.disks)) == 3


def test_disk_resize(autodelete_blank_disk, autodelete_regional_blank_disk):
    resize_disk(PROJECT, autodelete_blank_disk.self_link, 22)
    resize_disk(PROJECT, autodelete_regional_blank_disk.self_link, 23)

    disk_client = compute_v1.DisksClient()
    regional_disk_client = compute_v1.RegionDisksClient()
    assert disk_client.get(project=PROJECT, zone=ZONE, disk=autodelete_blank_disk.name).size_gb == 22
    assert regional_disk_client.get(project=PROJECT, region=REGION, disk=autodelete_regional_blank_disk.name).size_gb == 23



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# This is an ingredient file. It is not meant to be run directly. Check the samples/snippets
# folder for complete code samples that are ready to be used.
# Disabling flake8 for the ingredients file, as it would fail F821 - undefined name check.
# flake8: noqa
from __future__ import annotations


from google.cloud import compute_v1


# <INGREDIENT create_route>
def create_route(project_id: str, network: str, route_name: str, destination_range: str, *,
                 next_hop_gateway: str | None = None,
                 next_hop_ip: str | None = None, next_hop_instance: str | None = None,
                 next_hop_vpn_tunnel: str | None = None, next_hop_ilb: str | None = None) -> compute_v1.Route:
    """
    Create a new route in selected network by providing a destination and next hop name.

    Note: The set of {next_hop_gateway, next_hop_ip, next_hop_instance, next_hop_vpn_tunnel,
        next_hop_ilb} is exclusive, you and only specify one of those parameters.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        network: name of the network the route will be created in. Available name formats:
            * https://www.googleapis.com/compute/v1/projects/{project_id}/global/networks/{network}
            * projects/{project_id}/global/networks/{network}
            * global/networks/{network}
        route_name: name of the new route.
        destination_range: range of destination IPs this route should be applied to. E.g. 10.0.0.0/16.
        next_hop_gateway: name of the gateway the traffic should be directed to.
        next_hop_ip: IP address the traffic should be directed to.
        next_hop_instance: name of the instance the traffic should be directed to. Name format:
            "projects/{project}/zones/{zone}/instances/{instance_name}"
        next_hop_vpn_tunnel: name of the VPN tunnel the traffic should be directed to. Name format:
            "projects/{project}/regions/{region}/vpnTunnels/{vpn_tunnel_name}"
        next_hop_ilb: name of a forwarding rule of the Internal Load Balancer the traffic
            should be directed to. Name format:
            "projects/{project}/regions/{region}/forwardingRules/{forwarding_rule_region}"

    Returns:
        A new compute_v1.Route object.
    """
    excl_args = {next_hop_instance, next_hop_ilb, next_hop_vpn_tunnel, next_hop_gateway, next_hop_ip}
    args_set = sum(1 if arg is not None else 0 for arg in excl_args)

    if args_set != 1:
        raise RuntimeError("You must specify exactly one next_hop_* parameter.")

    route = compute_v1.Route()
    route.name = route_name
    route.network = network
    route.dest_range = destination_range

    if next_hop_gateway:
        route.next_hop_gateway = next_hop_gateway
    elif next_hop_ip:
        route.next_hop_ip = next_hop_ip
    elif next_hop_instance:
        route.next_hop_instance = next_hop_instance
    elif next_hop_vpn_tunnel:
        route.next_hop_vpn_tunnel = next_hop_vpn_tunnel
    elif next_hop_ilb:
        route.next_hop_ilb = next_hop_ilb

    route_client = compute_v1.RoutesClient()
    operation = route_client.insert(project=project_id, route_resource=route)

    wait_for_extended_operation(operation, "route creation")

    return route_client.get(project=project_id, route=route_name)
# </INGREDIENT>



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# This is an ingredient file. It is not meant to be run directly. Check the samples/snippets 
# folder for complete code samples that are ready to be used.
# Disabling flake8 for the ingredients file, as it would fail F821 - undefined name check.
# flake8: noqa
from google.cloud import compute_v1


# <INGREDIENT get_usage_export_bucket>
def get_usage_export_bucket(project_id: str) -> compute_v1.UsageExportLocation:
    """
    Retrieve Compute Engine usage export bucket for the Cloud project.
    Replaces the empty value returned by the API with the default value used
    to generate report file names.

    Args:
        project_id: project ID or project number of the project to update.
    Returns:
        UsageExportLocation object describing the current usage export settings
        for project project_id.
    """
    projects_client = compute_v1.ProjectsClient()
    project_data = projects_client.get(project=project_id)

    uel = project_data.usage_export_location

    if not uel.bucket_name:
        # The usage reports are disabled.
        return uel

    if not uel.report_name_prefix:
        # Although the server sent the empty string value, the next usage report
        # generated with these settings still has the default prefix value
        # "usage_gce". (see https://cloud.google.com/compute/docs/reference/rest/v1/projects/get)
        print(
            "Report name prefix not set, replacing with default value of "
            "`usage_gce`."
        )
        uel.report_name_prefix = "usage_gce"
    return uel
# </INGREDIENT>




#  Copyright 2023 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# This is an ingredient file. It is not meant to be run directly. Check the samples/snippets
# folder for complete code samples that are ready to be used.
# Disabling flake8 for the ingredients file, as it would fail F821 - undefined name check.
# flake8: noqa
import re

from google.cloud import compute_v1


# <INGREDIENT resize_disk>
def resize_disk(project_id: str, disk_link: str, new_size_gb: int) -> None:
    """
    Resizes a persistent disk to a specified size in GB. After you resize the disk, you must
    also resize the file system so that the operating system can access the additional space.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        disk_link: a link to the disk you want to resize.
            This value uses the following format:
                * https://www.googleapis.com/compute/v1/projects/{project_name}/zones/{zone}/disks/{disk_name}
                * projects/{project_name}/zones/{zone}/disks/{disk_name}
                * projects/{project_name}/regions/{region}/disks/{disk_name}
        new_size_gb: the new size you want to set for the disk in gigabytes.
    """
    search_results = re.search(r"/projects/[\w_-]+/(?P<area_type>zones|regions)/"
                               r"(?P<area_name>[\w_-]+)/disks/(?P<disk_name>[\w_-]+)", disk_link)

    if search_results["area_type"] == "regions":
        disk_client = compute_v1.RegionDisksClient()
        request = compute_v1.ResizeRegionDiskRequest()
        request.region = search_results["area_name"]
        request.region_disks_resize_request_resource = compute_v1.RegionDisksResizeRequest()
        request.region_disks_resize_request_resource.size_gb = new_size_gb
    else:
        disk_client = compute_v1.DisksClient()
        request = compute_v1.ResizeDiskRequest()
        request.zone = search_results["area_name"]
        request.disks_resize_request_resource = compute_v1.DisksResizeRequest()
        request.disks_resize_request_resource.size_gb = new_size_gb

    request.disk = search_results["disk_name"]
    request.project = project_id

    operation = disk_client.resize(request)
    wait_for_extended_operation(operation, "disk resize")
# </INGREDIENT>



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# flake8: noqa


# This file is automatically generated. Please do not modify it directly.
# Find the relevant recipe file in the samples/recipes or samples/ingredients
# directory and apply your changes there.


# [START compute_images_create_from_snapshot]
from __future__ import annotations

from collections.abc import Iterable
import sys
from typing import Any

from google.api_core.extended_operation import ExtendedOperation
from google.cloud import compute_v1


def wait_for_extended_operation(
    operation: ExtendedOperation, verbose_name: str = "operation", timeout: int = 300
) -> Any:
    """
    Waits for the extended (long-running) operation to complete.

    If the operation is successful, it will return its result.
    If the operation ends with an error, an exception will be raised.
    If there were any warnings during the execution of the operation
    they will be printed to sys.stderr.

    Args:
        operation: a long-running operation you want to wait on.
        verbose_name: (optional) a more verbose name of the operation,
            used only during error and warning reporting.
        timeout: how long (in seconds) to wait for operation to finish.
            If None, wait indefinitely.

    Returns:
        Whatever the operation.result() returns.

    Raises:
        This method will raise the exception received from `operation.exception()`
        or RuntimeError if there is no exception set, but there is an `error_code`
        set for the `operation`.

        In case of an operation taking longer than `timeout` seconds to complete,
        a `concurrent.futures.TimeoutError` will be raised.
    """
    result = operation.result(timeout=timeout)

    if operation.error_code:
        print(
            f"Error during {verbose_name}: [Code: {operation.error_code}]: {operation.error_message}",
            file=sys.stderr,
            flush=True,
        )
        print(f"Operation ID: {operation.name}", file=sys.stderr, flush=True)
        raise operation.exception() or RuntimeError(operation.error_message)

    if operation.warnings:
        print(f"Warnings during {verbose_name}:\n", file=sys.stderr, flush=True)
        for warning in operation.warnings:
            print(f" - {warning.code}: {warning.message}", file=sys.stderr, flush=True)

    return result


def create_image_from_snapshot(
    project_id: str,
    source_snapshot_name: str,
    image_name: str,
    source_project_id: str | None = None,
    guest_os_features: Iterable[str] | None = None,
    storage_location: str | None = None,
) -> compute_v1.Image:
    """
    Creates an image based on a snapshot.

    Args:
        project_id: project ID or project number of the Cloud project you want to place your new image in.
        source_snapshot_name: name of the snapshot you want to use as a base of your image.
        image_name: name of the image you want to create.
        source_project_id: name of the project that hosts the source snapshot. If left unset, it's assumed to equal
            the `project_id`.
        guest_os_features: an iterable collection of guest features you want to enable for the bootable image.
            Learn more about Guest OS features here:
            https://cloud.google.com/compute/docs/images/create-delete-deprecate-private-images#guest-os-features
        storage_location: the storage location of your image. For example, specify "us" to store the image in the
            `us` multi-region, or "us-central1" to store it in the `us-central1` region. If you do not make a selection,
             Compute Engine stores the image in the multi-region closest to your image's source location.

    Returns:
        An Image object.
    """
    if source_project_id is None:
        source_project_id = project_id

    snapshot_client = compute_v1.SnapshotsClient()
    image_client = compute_v1.ImagesClient()
    src_snapshot = snapshot_client.get(
        project=source_project_id, snapshot=source_snapshot_name
    )

    image = compute_v1.Image()
    image.name = image_name
    image.source_snapshot = src_snapshot.self_link

    if storage_location:
        image.storage_locations = [storage_location]

    if guest_os_features:
        image.guest_os_features = [
            compute_v1.GuestOsFeature(type_=feature) for feature in guest_os_features
        ]

    operation = image_client.insert(project=project_id, image_resource=image)

    wait_for_extended_operation(operation, "image creation from snapshot")

    return image_client.get(project=project_id, image=image_name)


# [END compute_images_create_from_snapshot]



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# This is an ingredient file. It is not meant to be run directly. Check the samples/snippets 
# folder for complete code samples that are ready to be used.
# Disabling flake8 for the ingredients file, as it would fail F821 - undefined name check.
# flake8: noqa

from google.cloud import compute_v1


# <INGREDIENT create_template_with_subnet>
def create_template_with_subnet(
    project_id: str, network: str, subnetwork: str, template_name: str
) -> compute_v1.InstanceTemplate:
    """
    Create an instance template that uses a provided subnet.

    Args:
        project_id: project ID or project number of the Cloud project you use.
        network: the network to be used in the new template. This value uses
            the following format: "projects/{project}/global/networks/{network}"
        subnetwork: the subnetwork to be used in the new template. This value
            uses the following format: "projects/{project}/regions/{region}/subnetworks/{subnetwork}"
        template_name: name of the new template to create.

    Returns:
        InstanceTemplate object that represents the new instance template.
    """
    # The template describes the size and source image of the book disk to
    # attach to the instance.
    disk = compute_v1.AttachedDisk()
    initialize_params = compute_v1.AttachedDiskInitializeParams()
    initialize_params.source_image = (
        "projects/debian-cloud/global/images/family/debian-11"
    )
    initialize_params.disk_size_gb = 250
    disk.initialize_params = initialize_params
    disk.auto_delete = True
    disk.boot = True

    template = compute_v1.InstanceTemplate()
    template.name = template_name
    template.properties = compute_v1.InstanceProperties()
    template.properties.disks = [disk]
    template.properties.machine_type = "e2-standard-4"

    # The template connects the instance to the specified network and subnetwork.
    network_interface = compute_v1.NetworkInterface()
    network_interface.network = network
    network_interface.subnetwork = subnetwork
    template.properties.network_interfaces = [network_interface]

    template_client = compute_v1.InstanceTemplatesClient()
    operation = template_client.insert(
        project=project_id, instance_template_resource=template
    )
    wait_for_extended_operation(operation, "instance template creation")

    return template_client.get(project=project_id, instance_template=template_name)
# </INGREDIENT>



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import uuid

import google.auth
import pytest

from ..disks.create_from_image import create_disk_from_image
from ..disks.delete import delete_disk
from ..images.get import get_image_from_family
from ..snapshots.create import create_snapshot
from ..snapshots.delete import delete_snapshot
from ..snapshots.get import get_snapshot
from ..snapshots.list import list_snapshots

PROJECT = google.auth.default()[1]
ZONE = 'europe-west1-c'


@pytest.fixture
def test_disk():
    debian_image = get_image_from_family('debian-cloud', 'debian-11')
    test_disk_name = "test-disk-" + uuid.uuid4().hex[:10]

    disk_type = f"zones/{ZONE}/diskTypes/pd-standard"

    disk = create_disk_from_image(PROJECT, ZONE, test_disk_name, disk_type, 20, debian_image.self_link)

    yield disk

    delete_disk(PROJECT, ZONE, test_disk_name)


def test_snapshot_create_delete(test_disk):
    snapshot_name = "test-snapshot-" + uuid.uuid4().hex[:10]
    snapshot = create_snapshot(PROJECT, test_disk.name, snapshot_name, zone=ZONE)
    assert snapshot.name == snapshot_name
    assert snapshot.source_disk == test_disk.self_link
    for i_snapshot in list_snapshots(PROJECT):
        if i_snapshot.name == snapshot_name:
            break
    else:
        pytest.fail("Couldn't find the created snapshot on snapshot list.")

    snapshot_get = get_snapshot(PROJECT, snapshot_name)
    assert snapshot_get.name == snapshot_name
    assert snapshot_get.disk_size_gb == snapshot.disk_size_gb
    assert snapshot_get.self_link == snapshot.self_link

    delete_snapshot(PROJECT, snapshot_name)
    for i_snapshot in list_snapshots(PROJECT):
        if i_snapshot.name == snapshot_name:
            pytest.fail("Test snapshot found on snapshot list, while it should already be gone.")



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


# This is an ingredient file. It is not meant to be run directly. Check the samples/snippets
# folder for complete code samples that are ready to be used.
# Disabling flake8 for the ingredients file, as it would fail F821 - undefined name check.
# flake8: noqa
from google.cloud import compute_v1


# <INGREDIENT set_deprecation_status>
def set_deprecation_status(project_id: str, image_name: str, status: compute_v1.DeprecationStatus.State) -> None:
    """
    Modify the deprecation status of an image.

    Note: Image objects by default don't have the `deprecated` attribute at all unless it's set.

    Args:
        project_id: project ID or project number of the Cloud project that hosts the image.
        image_name: name of the image you want to modify
        status: the status you want to set for the image. Available values are available in
            `compute_v1.DeprecationStatus.State` enum. Learn more about image deprecation statuses:
            https://cloud.google.com/compute/docs/images/create-delete-deprecate-private-images#deprecation-states
    """
    image_client = compute_v1.ImagesClient()
    deprecation_status = compute_v1.DeprecationStatus()
    deprecation_status.state = status.name
    operation = image_client.deprecate(project=project_id, image=image_name,
                                       deprecation_status_resource=deprecation_status)

    wait_for_extended_operation(operation, "changing deprecation state of an image")
# </INGREDIENT>



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# flake8: noqa


# This file is automatically generated. Please do not modify it directly.
# Find the relevant recipe file in the samples/recipes or samples/ingredients
# directory and apply your changes there.


# [START compute_route_list]
from __future__ import annotations

from collections.abc import Iterable

from google.cloud import compute_v1


def list_routes(
    project_id: str,
) -> Iterable[compute_v1.Route]:
    """
    Lists routes in project.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.

    Returns:
        An iterable collection of routes found in given project.
    """

    route_client = compute_v1.RoutesClient()
    return route_client.list(project=project_id)


# [END compute_route_list]



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# This is an ingredient file. It is not meant to be run directly. Check the samples/snippets 
# folder for complete code samples that are ready to be used.
# Disabling flake8 for the ingredients file, as it would fail F821 - undefined name check.
# flake8: noqa
from google.cloud import compute_v1


# <INGREDIENT create_with_subnet>
def create_with_subnet(
    project_id: str, zone: str, instance_name: str, network_link: str, subnet_link: str
) -> compute_v1.Instance:
    """
    Create a new VM instance with Debian 10 operating system in specified network and subnetwork.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        zone: name of the zone to create the instance in. For example: "us-west3-b"
        instance_name: name of the new virtual machine (VM) instance.
        network_link: name of the network you want the new instance to use.
            For example: "global/networks/default" represents the network
            named "default", which is created automatically for each project.
        subnetwork_link: name of the subnetwork you want the new instance to use.
            This value uses the following format:
            "regions/{region}/subnetworks/{subnetwork_name}"

    Returns:
        Instance object.
    """
    newest_debian = get_image_from_family(
        project="debian-cloud", family="debian-10"
    )
    disk_type = f"zones/{zone}/diskTypes/pd-standard"
    disks = [disk_from_image(disk_type, 10, True, newest_debian.self_link)]
    instance = create_instance(
        project_id,
        zone,
        instance_name,
        disks,
        network_link=network_link,
        subnetwork_link=subnet_link,
    )
    return instance
# </INGREDIENT>



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from argparse import Namespace
import glob
from pathlib import Path
import tempfile

import pytest

from . import sgs

FIXTURE_INGREDIENTS = Path("sgs_test_fixtures/ingredients")
FIXTURE_RECIPES = Path("sgs_test_fixtures/recipes")
FIXTURE_OUTPUT = Path("sgs_test_fixtures/output")


def test_sgs_generate():
    with tempfile.TemporaryDirectory() as tmp_dir:
        args = Namespace(output_dir=tmp_dir)
        sgs.generate(args, FIXTURE_INGREDIENTS.absolute(), FIXTURE_RECIPES.absolute())
        for test_file in map(Path, glob.glob(f"{tmp_dir}/**")):
            match_file = FIXTURE_OUTPUT / test_file.relative_to(tmp_dir)
            assert test_file.read_bytes() == match_file.read_bytes()


def test_snippets_freshness():
    """
    Make sure that the snippets/ folder is up-to-date and matches
    ingredients/ and recipes/. This test will generate SGS output
    in a temporary directory and compare it to the content of
    snippets/ folder.
    """
    with tempfile.TemporaryDirectory() as tmp_dir:
        args = Namespace(output_dir=tmp_dir)
        sgs.generate(args, Path("ingredients/").absolute(), Path("recipes/").absolute())
        print(list(map(Path, glob.glob(f"{tmp_dir}/**"))))
        for test_file in map(Path, glob.glob(f"{tmp_dir}/**", recursive=True)):
            match_file = Path("snippets/") / test_file.relative_to(tmp_dir)
            if test_file.is_file():
                if test_file.read_bytes() != match_file.read_bytes():
                    pytest.fail(
                        f"This test fails because file {match_file} seems to be outdated. Please run "
                        f"`python sgs.py generate` to update your snippets."
                    )
            elif test_file.is_dir():
                assert match_file.is_dir()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
import uuid

import google.api_core.exceptions
import google.auth
from google.cloud import compute_v1
import pytest

from ..firewall.create import create_firewall_rule
from ..firewall.delete import delete_firewall_rule
from ..firewall.main import get_firewall_rule
from ..firewall.patch import patch_firewall_priority

PROJECT = google.auth.default()[1]


@pytest.fixture
def firewall_rule():
    firewall_rule = compute_v1.Firewall()
    firewall_rule.name = "firewall-sample-test" + uuid.uuid4().hex[:10]
    firewall_rule.direction = "INGRESS"
    allowed_ports = compute_v1.Allowed()
    allowed_ports.I_p_protocol = "tcp"
    allowed_ports.ports = ["80"]
    firewall_rule.allowed = [allowed_ports]
    firewall_rule.source_ranges = ["0.0.0.0/0"]
    firewall_rule.network = "global/networks/default"
    firewall_rule.description = "Rule generated by Python sample test fixture."
    firewall_rule.target_tags = ["web"]

    firewall_client = compute_v1.FirewallsClient()
    op = firewall_client.insert_unary(project=PROJECT, firewall_resource=firewall_rule)

    op_client = compute_v1.GlobalOperationsClient()
    op_client.wait(project=PROJECT, operation=op.name)

    yield firewall_client.get(project=PROJECT, firewall=firewall_rule.name)

    try:
        op = firewall_client.delete_unary(project=PROJECT, firewall=firewall_rule.name)
        op_client.wait(project=PROJECT, operation=op.name)
    except google.api_core.exceptions.BadRequest as err:
        if err.code == 400 and "is not ready" in err.message:
            # This means GCE enforcer has already deleted that rule.
            pass
        else:
            raise err


@pytest.fixture
def autodelete_firewall_name():
    """
    Provide a name for a firewall rule and then delete the rule.
    """
    rule_name = "firewall-sample-test-" + uuid.uuid4().hex[:10]
    yield rule_name
    try:
        delete_firewall_rule(PROJECT, rule_name)
    except google.api_core.exceptions.BadRequest as err:
        if err.code == 400 and "is not ready" in err.message:
            # We can ignore this, this is most likely GCE Enforcer removing the rule before us.
            pass
        else:
            # Something else went wrong, let's escalate it.
            raise err


def test_create(autodelete_firewall_name):
    create_firewall_rule(PROJECT, autodelete_firewall_name)
    rule = get_firewall_rule(PROJECT, autodelete_firewall_name)
    assert rule.name == autodelete_firewall_name
    assert "web" in rule.target_tags


def test_patch_rule(firewall_rule):
    fw_client = compute_v1.FirewallsClient()
    assert firewall_rule.priority == 1000
    try:
        patch_firewall_priority(PROJECT, firewall_rule.name, 500)
    except google.api_core.exceptions.BadRequest as err:
        if err.code == 400 and "is not ready" in err.message:
            # We can ignore this, this is most likely GCE Enforcer removing the rule before us.
            return
        else:
            # Something else went wrong, let's escalate it.
            raise err
    time.sleep(2)
    updated_firewall_rule = fw_client.get(project=PROJECT, firewall=firewall_rule.name)
    assert updated_firewall_rule.priority == 500



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# This is an ingredient file. It is not meant to be run directly. Check the samples/snippets
# folder for complete code samples that are ready to be used.
# Disabling flake8 for the ingredients file, as it would fail F821 - undefined name check.
# flake8: noqa
from google.cloud import compute_v1


# <INGREDIENT create_disk_from_image>
def create_disk_from_image(
    project_id: str, zone: str, disk_name: str, disk_type: str, disk_size_gb: int, source_image: str
) -> compute_v1.Disk:
    """
    Creates a new disk in a project in given zone using an image as base.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        zone: name of the zone in which you want to create the disk.
        disk_name: name of the disk you want to create.
        disk_type: the type of disk you want to create. This value uses the following format:
            "zones/{zone}/diskTypes/(pd-standard|pd-ssd|pd-balanced|pd-extreme)".
            For example: "zones/us-west3-b/diskTypes/pd-ssd"
        disk_size_gb: size of the new disk in gigabytes
        source_image: source image to use when creating this disk. You must have read access to this disk. This
            can be one of the publicly available images or an image from one of your projects.
            This value uses the following format: "projects/{project_name}/global/images/{image_name}"

    Returns:
        An unattached Disk instance.
    """
    disk = compute_v1.Disk()
    disk.size_gb = disk_size_gb
    disk.name = disk_name
    disk.zone = zone
    disk.type_ = disk_type
    disk.source_image = source_image

    disk_client = compute_v1.DisksClient()
    operation = disk_client.insert(project=project_id, zone=zone, disk_resource=disk)

    wait_for_extended_operation(operation, "disk creation")

    return disk_client.get(project=project_id, zone=zone, disk=disk.name)
# </INGREDIENT>



from ast import List
from eco4cast.master_machine.weather_data_utils import (
    get_last_weather_data,
)
from eco4cast.master_machine.co2_model import CO2Model
import torch
import numpy as np
import datetime
from eco4cast.master_machine.electricitymaps_api import get_24h_history
from eco4cast.master_machine.utils import (
    codes_with_steps,
    code_names,
)
import pickle
import importlib.resources
from . import data_files
from eco4cast.master_machine.utils import code_names
from tqdm import tqdm


class CO2Predictor:
    """
    Class that uses TCNModel to predict co2 emission
    """

    def __init__(
        self,
        electricity_maps_api_key: str,
        checkpoint_file_path=None,
    ) -> None:
        """
        Args:
            electricity_maps_api_key : api key to access weather in electricity_maps regions
            (Optional) checkpoint_file_path (str) : path to file with torch model. If not provided - default model is used
        """

        self.lookback_window = 24  # Model-specific parameter

        self.data_path = importlib.resources.files(data_files)
        if checkpoint_file_path is None:
            checkpoint_file_path = self.data_path / "co2_model.ckpt"
        self.load_model(checkpoint_file_path=checkpoint_file_path)
        self.electricity_maps_api_key = electricity_maps_api_key

    def load_model(
        self,
        checkpoint_file_path,
    ):
        """
        This function loads model and country points for predicting time intervals.
        """
        self.predict_model = CO2Model(
            tcn_hparams={
                "num_inputs": 23 * 38,
                "num_channels": [512, 256, 256, 256, 256],
                "kernel_size": 3,
                "dropout": 0.0,
            },
            predict_window=24,
            optimizer_name="Adam",
            optimizer_hparams={"lr": 1e-3, "weight_decay": 1e-4},
        )
        self.predict_model = self.predict_model.load_from_checkpoint(
            checkpoint_file_path, map_location=torch.device("cpu")
        )
        self.predict_model = self.predict_model.to("cpu").eval()

        self.zones_with_steps = codes_with_steps

        with open(self.data_path / "country_points.pickle", "rb") as f:
            self.country_points = pickle.load(f)

    def predict_co2(self):
        """
        This function predicts co2 emission using model (firstly, it loads weather and emission data).
        """

        batch = []
        for code, points in tqdm(self.country_points):
            point_weather_matrices = []
            zone_emission = get_24h_history(code, self.electricity_maps_api_key)
            # zone_emission = [0] * 24 # For testing purposes
            for longitude, latitude in points:
                point_weather = get_last_weather_data(
                    latitude=latitude,
                    longitude=longitude,
                    lookback_days=(self.lookback_window // 24) + 1,
                ).iloc[-self.lookback_window :]
                point_weather = point_weather.drop(columns=["weathercode", "time"])
                point_weather["emission"] = zone_emission
                point_weather["longitude"] = longitude
                point_weather["latitude"] = latitude
                point_weather = point_weather.to_numpy()
                point_weather_matrices.append(point_weather)

            for _ in range(len(points), 38):
                point_weather_matrices.append(np.zeros_like(point_weather))

            point_weather_matrices = np.array(point_weather_matrices)
            batch.append(point_weather_matrices)

        batch = np.array(batch)
        batch = torch.tensor(batch.astype(np.float32))
        batch = batch.permute((0, 2, 3, 1))

        self.emission_forecast = self.predict_model.predict_step((batch, 0, 0), 0)
        self.emission_forecast = self.emission_forecast.detach().cpu().numpy()

        return self.emission_forecast


class IntervalGenerator:
    """
    Class that uses CO2Predictor's forecast to generate intervals that satisfy conditions
    (emission in interval less than threshold)
    """

    def __init__(
        self,
        zone_names=code_names,
        max_emission_value=180,
        co2_delta_to_move=30,
        min_interval_size=1,
        max_window_size=3,
        exclude_zones=None,
        include_zones=None,
    ) -> None:
        self.zone_names = dict(zip(list(range(len(zone_names))), zone_names))
        self.zone_names[-1] = -1
        self.zone_to_id = dict(zip(zone_names, list(range(len(zone_names)))))
        self.max_emission_value = max_emission_value
        self.co2_delta_to_move = co2_delta_to_move
        self.min_interval_size = min_interval_size
        self.max_window_size = max_window_size
        self.exclude_zones = exclude_zones
        self.include_zones = include_zones

    def generate_intervals(
        self,
        forecasts,
        current_zone_idx=0,
    ):
        """
        This function generates training intervals.
        Uses co2_predictor's forecasted co2 emission to build intervals with minimum total emission.
        """

        assert len(forecasts) == len(self.zone_names) - 1

        forecasts = np.copy(forecasts)
        

        if self.include_zones is not None and len(self.include_zones) > 0:
            self.exclude_zones = [
                k for k in self.zone_to_id.keys() if k not in self.include_zones
            ]

        if self.exclude_zones is not None and len(self.exclude_zones) > 0:
            for z in self.exclude_zones:
                forecasts[self.zone_to_id[z]] = 1e9

        time_slots = np.zeros((len(forecasts), 24), dtype=int)

        for forecast_id, forecast in enumerate(forecasts):
            for window_size in range(self.max_window_size):
                for i in range(len(forecast)):
                    co2_mean = forecast[
                        i : i + self.min_interval_size + window_size
                    ].mean()
                    if co2_mean < self.max_emission_value:
                        time_slots[
                            forecast_id,
                            i : i + self.min_interval_size + window_size - 1,
                        ] = 1

        time_slots_vms = np.ones((24), dtype=int) * -1

        i = 0
        current_vm_idx = current_zone_idx

        while i + self.min_interval_size <= 24:
            if time_slots[:, i].sum() == 0:
                i += 1
                continue

            possible_vms = time_slots[:, i : i + self.min_interval_size].all(1)
            if possible_vms.sum() == 1:
                time_slots_vms[
                    i : i + self.min_interval_size
                ] = current_vm_idx = possible_vms.argmax()
                i += self.min_interval_size
            else:
                co2_mean = forecasts[:, i : i + self.min_interval_size].mean(axis=1)
                co2_mean += (~possible_vms) * 1e9
                co2_mean[current_vm_idx] -= self.co2_delta_to_move
                min_co2_idx = current_vm_idx = co2_mean.argmin()
                time_slots_vms[i : i + self.min_interval_size] = min_co2_idx
                i += self.min_interval_size

        current_vm = -1
        intervals = {}
        for time, vm in enumerate(time_slots_vms):
            if current_vm == -1:
                current_vm = vm
                start_time = time

            if vm != current_vm:
                end_time = time
                if current_vm in intervals.keys():
                    intervals[current_vm].append((start_time, end_time))
                else:
                    intervals[current_vm] = [(start_time, end_time)]
                current_vm = vm
                start_time = time

        end_time = time
        if current_vm in intervals.keys():
            intervals[current_vm].append((start_time, end_time))
        else:
            intervals[current_vm] = [(start_time, end_time)]

        if -1 in intervals.keys():
            del intervals[-1]

        ordered_intervals = [
            (k, intervals[k][i])
            for k in intervals.keys()
            for i in range(len(intervals[k]))
        ]
        ordered_intervals = sorted(ordered_intervals, key=lambda x: x[1][0])

        if len(ordered_intervals) == 0:
            return [], current_zone_idx
        
        formatted_intervals = [(ordered_intervals[0][0], [ordered_intervals[0][1]])]

        for vm_idx, interval in ordered_intervals[1:]:
            if formatted_intervals[-1][0] == vm_idx:
                formatted_intervals[-1][1].append(interval)
            else:
                formatted_intervals.append((vm_idx, [interval]))

        formatted_intervals

        datetime_intervals = [
            (
                self.zone_names[vm_idx],
                [
                    (
                        datetime.datetime.now(datetime.timezone.utc).replace(
                            minute=0, second=0, microsecond=0
                        )
                        + datetime.timedelta(hours=int(jstart)),
                        datetime.datetime.now(datetime.timezone.utc).replace(
                            minute=0, second=0, microsecond=0
                        )
                        + datetime.timedelta(hours=int(jend)),
                        forecasts[vm_idx, jstart : jend].mean()
                    ) 
                    for jstart, jend in vm_intervals
                ],
            )
            for vm_idx, vm_intervals in formatted_intervals
        ]

        return datetime_intervals, [
            vm_idx for vm_idx, vm_intervals in formatted_intervals
        ]



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# flake8: noqa
from __future__ import annotations

from collections.abc import Iterable
import uuid

from google.cloud import compute_v1


# <INGREDIENT bulk_insert_instance>
def bulk_insert_instance(project_id: str, zone: str, template: compute_v1.InstanceTemplate,
                         count: int, name_pattern: str, min_count: int | None = None,
                         labels: dict | None = None) -> Iterable[compute_v1.Instance]:
    """
    Create multiple VMs based on an Instance Template. The newly created instances will
    be returned as a list and will share a label with key `bulk_batch` and a random
    value.

    If the bulk insert operation fails and the requested number of instances can't be created,
    and more than min_count instances are created, then those instances can be found using
    the `bulk_batch` label with value attached to the raised exception in bulk_batch_id
    attribute. So, you can use the following filter: f"label.bulk_batch={err.bulk_batch_id}"
    when listing instances in a zone to get the instances that were successfully created.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        zone: name of the zone to create the instance in. For example: "us-west3-b"
        template: an Instance Template to be used for creation of the new VMs.
        name_pattern: The string pattern used for the names of the VMs. The pattern
            must contain one continuous sequence of placeholder hash characters (#)
            with each character corresponding to one digit of the generated instance
            name. Example: a name_pattern of inst-#### generates instance names such
            as inst-0001 and inst-0002. If existing instances in the same project and
            zone have names that match the name pattern then the generated instance
            numbers start after the biggest existing number. For example, if there
            exists an instance with name inst-0050, then instance names generated
            using the pattern inst-#### begin with inst-0051. The name pattern
            placeholder #...# can contain up to 18 characters.
        count: The maximum number of instances to create.
        min_count (optional): The minimum number of instances to create. If no min_count is
            specified then count is used as the default value. If min_count instances
            cannot be created, then no instances will be created and instances already
            created will be deleted.
        labels (optional): A dictionary with labels to be added to the new VMs.
    """
    bulk_insert_resource = compute_v1.BulkInsertInstanceResource()
    bulk_insert_resource.source_instance_template = template.self_link
    bulk_insert_resource.count = count
    bulk_insert_resource.min_count = min_count or count
    bulk_insert_resource.name_pattern = name_pattern

    if not labels:
        labels = {}

    labels['bulk_batch'] = uuid.uuid4().hex
    instance_prop = compute_v1.InstanceProperties()
    instance_prop.labels = labels
    bulk_insert_resource.instance_properties = instance_prop

    bulk_insert_request = compute_v1.BulkInsertInstanceRequest()
    bulk_insert_request.bulk_insert_instance_resource_resource = bulk_insert_resource
    bulk_insert_request.project = project_id
    bulk_insert_request.zone = zone

    client = compute_v1.InstancesClient()
    operation = client.bulk_insert(bulk_insert_request)

    try:
        wait_for_extended_operation(operation, "bulk instance creation")
    except Exception as err:
        err.bulk_batch_id = labels['bulk_batch']
        raise err

    list_req = compute_v1.ListInstancesRequest()
    list_req.project = project_id
    list_req.zone = zone
    list_req.filter = " AND ".join(f"labels.{key}:{value}" for key, value in labels.items())
    return client.list(list_req)
# </INGREDIENT>



#  Copyright 2023 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# This is an ingredient file. It is not meant to be run directly. Check the samples/snippets
# folder for complete code samples that are ready to be used.
# Disabling flake8 for the ingredients file, as it would fail F821 - undefined name check.
# flake8: noqa

from google.cloud import compute_v1


# <INGREDIENT change_machine_type>
def change_machine_type(
        project_id: str,
        zone: str,
        instance_name: str,
        new_machine_type: str
) -> None:
    """
    Changes the machine type of VM. The VM needs to be in the 'TERMINATED' state for this operation to be successful.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        zone: name of the zone your instance belongs to.
        instance_name: name of the VM you want to modify.
        new_machine_type: the new machine type you want to use for the VM.
            For example: `e2-standard-8`, `e2-custom-4-2048` or `m1-ultramem-40`
            More about machine types: https://cloud.google.com/compute/docs/machine-resource
    """
    client = compute_v1.InstancesClient()
    instance = client.get(project=project_id, zone=zone, instance=instance_name)

    if instance.status != compute_v1.Instance.Status.TERMINATED.name:
        raise RuntimeError(f"Only machines in TERMINATED state can have their machine type changed. "
                           f"{instance.name} is in {instance.status}({instance.status_message}) state.")

    machine_type = compute_v1.InstancesSetMachineTypeRequest()
    machine_type.machine_type = f"projects/{project_id}/zones/{zone}/machineTypes/{new_machine_type}"
    operation = client.set_machine_type(
        project=project_id,
        zone=zone,
        instance=instance_name,
        instances_set_machine_type_request_resource=machine_type,
    )

    wait_for_extended_operation(operation, "changing machine type")
# </INGREDIENT>



import paramiko
import os
import datetime
from eco4cast.master_machine.interval_predictor import (
    CO2Predictor,
    IntervalGenerator,
)
from time import sleep
from eco4cast.master_machine.compute.client_library.snippets.instances.stop import (
    stop_instance,
)
from eco4cast.master_machine.google_cloud_vm_moving import google_cloud_move_vm
from eco4cast.master_machine.utils import codes_to_gcloud_zones
import time
import pandas as pd


class Controller:
    def __init__(
        self,
        credentials_path,
        ssh_username,
        ssh_python_path,
        ssh_vm_main_path,
        current_vm_ip,
        ssh_port,
        current_zone,
        current_instance_name,
        project_id,
        intervals_prediction_period,
        co2_predictor: CO2Predictor,
        interval_generator: IntervalGenerator,
    ):
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = str(credentials_path)

        self.load_states = False
        self.last_prediction_time = datetime.datetime.now()
        self.co2_predictor = co2_predictor
        print("Gathering info and predicting CO2 ... This may take a while ...")
        self.co2_forecast = self.co2_predictor.predict_co2()
        print("CO2 predicted")
        self.interval_generator = interval_generator

        (
            self.predicted_intervals,
            self.zone_indices,
        ) = self.interval_generator.generate_intervals(
            forecasts=self.co2_forecast,
        )

        self.ssh_username = ssh_username
        self.ssh_python_path = ssh_python_path
        self.ssh_vm_main_path = ssh_vm_main_path
        self.current_vm_ip = current_vm_ip
        self.ssh_port = ssh_port
        self.current_zone = current_zone
        self.current_instance_name = current_instance_name
        self.project_id = project_id
        self.intervals_prediction_period = intervals_prediction_period

        # FOR TESTING
        # self.predicted_intervals = [
        #     # (
        #     #     "FR",
        #     #     [
        #     #         (
        #     #             datetime.datetime(2023, 7, 13, 5, 0, tzinfo=datetime.timezone.utc),
        #     #             datetime.datetime(2023, 7, 13, 5, 10, tzinfo=datetime.timezone.utc),
        #     #         )
        #     #     ],
        #     # ),
        #     (
        #         "CA-QC",
        #         [
        #             (
        #                 datetime.datetime(2023, 7, 13, 5, 15, tzinfo=datetime.timezone.utc),
        #                 datetime.datetime(2023, 7, 14, 8, 0, tzinfo=datetime.timezone.utc),
        #             )
        #         ],
        #     )
        # ]

    def start_training(self):
        interval_idx = 0
        shutting = False
        while interval_idx < len(self.predicted_intervals) and not shutting:
            code, time_intervals = self.predicted_intervals[interval_idx]
            zone_idx = self.zone_indices[interval_idx]
            zone = codes_to_gcloud_zones[code]
            if zone != self.current_zone:
                print("Moving VM")
                start_moving_time = time.time()
                new_instance = google_cloud_move_vm(
                    current_zone=self.current_zone,
                    current_instance_name=self.current_instance_name,
                    project_id=self.project_id,
                    new_zone=zone,
                )
                print(
                    f"Moving completed in {int(time.time()-start_moving_time)} seconds"
                )
                self.current_vm_ip = (
                    new_instance.network_interfaces[0].access_configs[0].nat_i_p
                )
                self.current_zone = zone
                sleep(60)  # To let machine start

            # Scheduling job 30 seconds before first start_time
            printed = False
            while datetime.datetime.now(datetime.timezone.utc) < time_intervals[0][
                0
            ] - datetime.timedelta(seconds=30):
                if not printed:
                    print(
                        f'Waiting till next interval start {time_intervals[0][0].strftime("%Y%m%d%H%M%S")}'
                    )
                time.sleep(1)

            argument_intervals = [
                (s.strftime("%Y%m%d%H%M%S"), e.strftime("%Y%m%d%H%M%S"))
                for s, e, _ in time_intervals
            ]

            argument_intervals = [item for t in argument_intervals for item in t]
            argument_intervals = ",".join(argument_intervals)

            command_arguments = f"--times {argument_intervals}"
            if self.load_states:
                command_arguments += " --load_states"

            co2_means = [str(co2) for _, _, co2 in time_intervals]
            command_arguments += f' --co2_means {",".join(co2_means)}'

            print("SSH Starting")
            channel, stdout = self.__setup_ssh_execution(
                self.current_vm_ip,
                self.ssh_port,
                self.ssh_python_path,
                self.ssh_vm_main_path,
                username=self.ssh_username,
                command_arguments=command_arguments,
            )
            try:
                out = b""
                while True:
                    step = 500
                    out = out + stdout.read(step)
                    if out == b"":
                        break
                    try:
                        line = out.decode()
                        print(line, end="\r")
                        out = b""
                        if "Traceback" in line:
                            pass
                        if "End of training." in line:
                            print("Stopping master machine")
                            shutting = True
                            break
                    except UnicodeDecodeError:
                        out = out
            except KeyboardInterrupt:
                channel.send("\x03")
                print("Interrupted")
                shutting = True

            if not shutting:
                interval_idx += 1
                if (
                    datetime.datetime.now()
                    > self.last_prediction_time
                    + datetime.timedelta(seconds=self.intervals_prediction_period)
                ):
                    self.last_prediction_time = datetime.datetime.now()
                    co2_forecast = self.co2_predictor.predict_co2()
                    self.predicted_intervals = (
                        self.interval_generator.generate_intervals(
                            forecasts=co2_forecast,
                            current_machine=zone_idx,
                        )
                    )
                    interval_idx = 0
                self.load_states = True

        self.__calc_emission()

    def __calc_emission(self):
        sftp = self.ssh_client.open_sftp()
        path = "/".join(self.ssh_vm_main_path.split("/")[:-1]) + "/emission.csv"
        sftp.get(path, "emission.csv")
        emission_df = pd.read_csv("emission.csv")
        total_emission = emission_df["CO2_emissions(kg)"].sum() * 1000
        total_electricity = emission_df["power_consumption(kWh)"].sum()

        co2_average_intensity = self.co2_forecast.mean()
        co2_average_g = total_electricity * co2_average_intensity

        if total_emission / co2_average_g <= 1:
            text = [
                f"Your total CO2 emissions are {total_emission:.6f} g. ",
                f"{co2_average_g - total_emission:.6f} g were saved in comparison with average emission ({co2_average_g:.6f} g)",
            ]
        else:
            text = [
                f"Your total CO2 emissions are {total_emission:.6f} g. ",
                f"It is {co2_average_g - total_emission:.6f} g more than average emission ({co2_average_g:.6f} g)",
            ]

        width = 80
        print("+=" + "=" * width + "=+")
        for line in text:
            print(f"|| {line.center(width-2)} ||")
        print("+=" + "=" * (width) + "=+")

    def __setup_ssh_execution(
        self,
        ip,
        ssh_port,
        python_path,
        vm_main_path,
        username="scheduler",
        command=None,
        command_arguments="",
    ):
        self.ssh_client = paramiko.SSHClient()
        self.ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.ssh_client.connect(
            hostname=ip, port=ssh_port, username=username, timeout=10
        )

        vm_main_folder = "/".join(vm_main_path.split("/")[:-1])
        vm_main_file = vm_main_path.split("/")[-1]

        command = f"cd {vm_main_folder}; {python_path} {vm_main_file}"
        command += " " + command_arguments

        transport = self.ssh_client.get_transport()
        bufsize = -1
        channel = transport.open_session()
        channel.get_pty()
        channel.exec_command(command)
        # stdin = channel.makefile_stdin("wb", bufsize)
        stdout = channel.makefile("r", bufsize)
        # stderr = channel.makefile_stderr("r", bufsize)

        return channel, stdout

    def stop_training(self):
        stop_instance(
            project_id=self.project_id,
            zone=self.current_zone,
            instance_name=self.current_instance_name,
        )



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# flake8: noqa

# <IMPORTS/>


# <INGREDIENT create_firewall_rule />

# <INGREDIENT delete_firewall_rule />

# <INGREDIENT get_firewall_rule />

# <INGREDIENT list_firewall_rules />

# <INGREDIENT patch_firewall_priority />

if __name__ == "__main__":
    import google.auth
    import google.auth.exceptions

    try:
        default_project_id = google.auth.default()[1]
        print(f"Using project {default_project_id}.")
    except google.auth.exceptions.DefaultCredentialsError:
        print(
            "Please use `gcloud auth application-default login` "
            "or set GOOGLE_APPLICATION_CREDENTIALS to use this script."
        )
    else:
        import uuid

        rule_name = "firewall-sample-" + uuid.uuid4().hex[:10]
        print(f"Creating firewall rule {rule_name}...")
        # The rule will be created with default priority of 1000.
        create_firewall_rule(default_project_id, rule_name)
        try:
            print("Rule created:")
            print(get_firewall_rule(default_project_id, rule_name))
            print("Updating rule priority to 10...")
            patch_firewall_priority(default_project_id, rule_name, 10)
            print("Rule updated: ")
            print(get_firewall_rule(default_project_id, rule_name))
            print(f"Deleting rule {rule_name}...")
        finally:
            delete_firewall_rule(default_project_id, rule_name)
        print("Done.")



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# flake8: noqa


# This file is automatically generated. Please do not modify it directly.
# Find the relevant recipe file in the samples/recipes or samples/ingredients
# directory and apply your changes there.


# [START compute_windows_image_create]
# [START compute_images_create]
from __future__ import annotations

import sys
from typing import Any
import warnings

from google.api_core.extended_operation import ExtendedOperation
from google.cloud import compute_v1


def wait_for_extended_operation(
    operation: ExtendedOperation, verbose_name: str = "operation", timeout: int = 300
) -> Any:
    """
    Waits for the extended (long-running) operation to complete.

    If the operation is successful, it will return its result.
    If the operation ends with an error, an exception will be raised.
    If there were any warnings during the execution of the operation
    they will be printed to sys.stderr.

    Args:
        operation: a long-running operation you want to wait on.
        verbose_name: (optional) a more verbose name of the operation,
            used only during error and warning reporting.
        timeout: how long (in seconds) to wait for operation to finish.
            If None, wait indefinitely.

    Returns:
        Whatever the operation.result() returns.

    Raises:
        This method will raise the exception received from `operation.exception()`
        or RuntimeError if there is no exception set, but there is an `error_code`
        set for the `operation`.

        In case of an operation taking longer than `timeout` seconds to complete,
        a `concurrent.futures.TimeoutError` will be raised.
    """
    result = operation.result(timeout=timeout)

    if operation.error_code:
        print(
            f"Error during {verbose_name}: [Code: {operation.error_code}]: {operation.error_message}",
            file=sys.stderr,
            flush=True,
        )
        print(f"Operation ID: {operation.name}", file=sys.stderr, flush=True)
        raise operation.exception() or RuntimeError(operation.error_message)

    if operation.warnings:
        print(f"Warnings during {verbose_name}:\n", file=sys.stderr, flush=True)
        for warning in operation.warnings:
            print(f" - {warning.code}: {warning.message}", file=sys.stderr, flush=True)

    return result


STOPPED_MACHINE_STATUS = (
    compute_v1.Instance.Status.TERMINATED.name,
    compute_v1.Instance.Status.STOPPED.name,
)


def create_image_from_disk(
    project_id: str,
    zone: str,
    source_disk_name: str,
    image_name: str,
    storage_location: str | None = None,
    force_create: bool = False,
) -> compute_v1.Image:
    """
    Creates a new disk image.

    Args:
        project_id: project ID or project number of the Cloud project you use.
        zone: zone of the disk you copy from.
        source_disk_name: name of the source disk you copy from.
        image_name: name of the image you want to create.
        storage_location: storage location for the image. If the value is undefined,
            function will store the image in the multi-region closest to your image's
            source location.
        force_create: create the image even if the source disk is attached to a
            running instance.

    Returns:
        An Image object.
    """
    image_client = compute_v1.ImagesClient()
    disk_client = compute_v1.DisksClient()
    instance_client = compute_v1.InstancesClient()

    # Get source disk
    disk = disk_client.get(project=project_id, zone=zone, disk=source_disk_name)

    for disk_user in disk.users:
        instance = instance_client.get(
            project=project_id, zone=zone, instance=disk_user
        )
        if instance.status in STOPPED_MACHINE_STATUS:
            continue
        if not force_create:
            raise RuntimeError(
                f"Instance {disk_user} should be stopped. For Windows instances please "
                f"stop the instance using `GCESysprep` command. For Linux instances just "
                f"shut it down normally. You can supress this error and create an image of"
                f"the disk by setting `force_create` parameter to true (not recommended). \n"
                f"More information here: \n"
                f" * https://cloud.google.com/compute/docs/instances/windows/creating-windows-os-image#api \n"
                f" * https://cloud.google.com/compute/docs/images/create-delete-deprecate-private-images#prepare_instance_for_image"
            )
        else:
            warnings.warn(
                f"Warning: The `force_create` option may compromise the integrity of your image. "
                f"Stop the {disk_user} instance before you create the image if possible."
            )

    # Create image
    image = compute_v1.Image()
    image.source_disk = disk.self_link
    image.name = image_name
    if storage_location:
        image.storage_locations = [storage_location]

    operation = image_client.insert(project=project_id, image_resource=image)

    wait_for_extended_operation(operation, "image creation from disk")

    return image_client.get(project=project_id, image=image_name)


# [END compute_images_create]
# [END compute_windows_image_create]



#  Copyright 2022 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# flake8: noqa


# This file is automatically generated. Please do not modify it directly.
# Find the relevant recipe file in the samples/recipes or samples/ingredients
# directory and apply your changes there.


# [START compute_preemptible_check]
from google.cloud import compute_v1


def is_preemptible(project_id: str, zone: str, instance_name: str) -> bool:
    """
    Check if a given instance is preemptible or not.
    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        zone: name of the zone you want to use. For example: "us-west3-b"
        instance_name: name of the virtual machine to check.
    Returns:
        The preemptible status of the instance.
    """
    instance_client = compute_v1.InstancesClient()
    instance = instance_client.get(
        project=project_id, zone=zone, instance=instance_name
    )
    return instance.scheduling.preemptible


# [END compute_preemptible_check]



#  Copyright 2021 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import uuid

import google.auth
import pytest

from ..images.get import get_image_from_family
from ..instances.create import create_instance
from ..instances.create_start_instance.create_from_public_image import disk_from_image
from ..instances.custom_machine_types.create_shared_with_helper import (
    create_custom_shared_core_instance,
)
from ..instances.custom_machine_types.create_with_helper import create_custom_instance
from ..instances.custom_machine_types.helper_class import CustomMachineType
from ..instances.custom_machine_types.update_memory import (
    add_extended_memory_to_instance,
)
from ..instances.delete import delete_instance

PROJECT = google.auth.default()[1]
REGION = "us-central1"
INSTANCE_ZONE = "us-central1-b"


@pytest.fixture
def auto_delete_instance_name():
    instance_name = "test-instance-" + uuid.uuid4().hex[:10]
    yield instance_name
    delete_instance(PROJECT, INSTANCE_ZONE, instance_name)


@pytest.fixture
def instance():
    instance_name = "test-instance-" + uuid.uuid4().hex[:10]

    newest_debian = get_image_from_family(project="debian-cloud", family="debian-10")
    disk_type = f"zones/{INSTANCE_ZONE}/diskTypes/pd-standard"
    disks = [disk_from_image(disk_type, 10, True, newest_debian.self_link)]

    instance = create_instance(
        PROJECT, INSTANCE_ZONE, instance_name, disks, "n2-custom-8-10240"
    )
    yield instance
    delete_instance(PROJECT, INSTANCE_ZONE, instance_name)


def test_custom_instance_creation(auto_delete_instance_name):
    # Need to import CustomMachineType from this module, or the assertion will fail
    from ..instances.custom_machine_types.create_with_helper import CustomMachineType

    instance = create_custom_instance(
        PROJECT,
        INSTANCE_ZONE,
        auto_delete_instance_name,
        CustomMachineType.CPUSeries.E2,
        4,
        8192,
    )

    assert instance.name == auto_delete_instance_name
    assert instance.machine_type.endswith(
        f"zones/{INSTANCE_ZONE}/machineTypes/e2-custom-4-8192"
    )


def test_custom_shared_instance_creation(auto_delete_instance_name):
    # Need to import CustomMachineType from this module, or the assertion will fail
    from ..instances.custom_machine_types.create_shared_with_helper import (
        CustomMachineType,
    )

    instance = create_custom_shared_core_instance(
        PROJECT,
        INSTANCE_ZONE,
        auto_delete_instance_name,
        CustomMachineType.CPUSeries.E2_MICRO,
        2048,
    )

    assert instance.name == auto_delete_instance_name
    assert instance.machine_type.endswith(
        f"zones/{INSTANCE_ZONE}/machineTypes/e2-custom-micro-2048"
    )


def test_custom_machine_type_good():
    # N1
    cmt = CustomMachineType(INSTANCE_ZONE, CustomMachineType.CPUSeries.N1, 8192, 8)
    assert str(cmt) == f"zones/{INSTANCE_ZONE}/machineTypes/custom-8-8192"
    assert cmt.short_type_str() == "custom-8-8192"
    # N2
    cmt = CustomMachineType(INSTANCE_ZONE, CustomMachineType.CPUSeries.N2, 4096, 4)
    assert str(cmt) == f"zones/{INSTANCE_ZONE}/machineTypes/n2-custom-4-4096"
    assert cmt.short_type_str() == "n2-custom-4-4096"
    # N2D
    cmt = CustomMachineType(INSTANCE_ZONE, CustomMachineType.CPUSeries.N2D, 8192, 4)
    assert str(cmt) == f"zones/{INSTANCE_ZONE}/machineTypes/n2d-custom-4-8192"
    assert cmt.short_type_str() == "n2d-custom-4-8192"
    # E2
    cmt = CustomMachineType(INSTANCE_ZONE, CustomMachineType.CPUSeries.E2, 8192, 8)
    assert str(cmt) == f"zones/{INSTANCE_ZONE}/machineTypes/e2-custom-8-8192"
    assert cmt.short_type_str() == "e2-custom-8-8192"
    # E2 SMALL
    cmt = CustomMachineType(INSTANCE_ZONE, CustomMachineType.CPUSeries.E2_SMALL, 4096)
    assert str(cmt) == f"zones/{INSTANCE_ZONE}/machineTypes/e2-custom-small-4096"
    assert cmt.short_type_str() == "e2-custom-small-4096"
    # E2 MICRO
    cmt = CustomMachineType(INSTANCE_ZONE, CustomMachineType.CPUSeries.E2_MICRO, 2048)
    assert str(cmt) == f"zones/{INSTANCE_ZONE}/machineTypes/e2-custom-micro-2048"
    assert cmt.short_type_str() == "e2-custom-micro-2048"
    # E2 MEDIUM
    cmt = CustomMachineType(INSTANCE_ZONE, CustomMachineType.CPUSeries.E2_MEDIUM, 8192)
    assert str(cmt) == f"zones/{INSTANCE_ZONE}/machineTypes/e2-custom-medium-8192"
    assert cmt.short_type_str() == "e2-custom-medium-8192"


def test_custom_machine_type_bad_memory_256():
    try:
        CustomMachineType(INSTANCE_ZONE, CustomMachineType.CPUSeries.N1, 8194, 8)
    except RuntimeError as err:
        assert err.args[0] == "Requested memory must be a multiple of 256 MB."
    else:
        assert not "This test should have raised an exception!"


def test_custom_machine_type_ext_memory():
    cmt = CustomMachineType(INSTANCE_ZONE, CustomMachineType.CPUSeries.N2, 638720, 8)
    assert str(cmt) == f"zones/{INSTANCE_ZONE}/machineTypes/n2-custom-8-638720-ext"


def test_custom_machine_type_bad_cpu_count():
    try:
        CustomMachineType(INSTANCE_ZONE, CustomMachineType.CPUSeries.N2, 8194, 66)
    except RuntimeError as err:
        assert err.args[0].startswith(
            "Invalid number of cores requested. Allowed number of cores for"
        )
    else:
        assert not "This test should have raised an exception!"


def test_add_extended_memory_to_instance(instance):
    instance = add_extended_memory_to_instance(
        PROJECT, INSTANCE_ZONE, instance.name, 819200
    )
    assert instance.machine_type.endswith("819200-ext")


def test_from_str_creation():
    cmt = CustomMachineType.from_str(
        "https://www.googleapis.com/compute/v1/projects/diregapic-mestiv/zones/us-central1-b/machineTypes/e2-custom-4-8192"
    )
    assert cmt.zone == "us-central1-b"
    assert cmt.memory_mb == 8192
    assert cmt.extra_memory_used is False
    assert cmt.cpu_series is CustomMachineType.CPUSeries.E2
    assert cmt.core_count == 4

    cmt = CustomMachineType.from_str(
        "zones/europe-west4-b/machineTypes/n2-custom-8-81920-ext"
    )
    assert cmt.zone == "europe-west4-b"
    assert cmt.memory_mb == 81920
    assert cmt.extra_memory_used is True
    assert cmt.cpu_series is CustomMachineType.CPUSeries.N2
    assert cmt.core_count == 8

    cmt = CustomMachineType.from_str(
        "zones/europe-west4-b/machineTypes/e2-custom-small-4096"
    )
    assert cmt.zone == "europe-west4-b"
    assert cmt.memory_mb == 4096
    assert cmt.extra_memory_used is False
    assert cmt.cpu_series == CustomMachineType.CPUSeries.E2_SMALL
    assert cmt.core_count == 2

    cmt = CustomMachineType.from_str(
        "zones/europe-west2-b/machineTypes/custom-2-2048"
    )
    assert cmt.zone == "europe-west2-b"
    assert cmt.memory_mb == 2048
    assert cmt.extra_memory_used is False
    assert cmt.cpu_series is CustomMachineType.CPUSeries.N1
    assert cmt.core_count == 2

    try:
        CustomMachineType.from_str(
            "zones/europe-west2-b/machineTypes/n8-custom-2-1024"
        )
    except RuntimeError as err:
        assert err.args[0] == "Unknown CPU series."
    else:
        assert not "This was supposed to raise a RuntimeError."

    cmt = CustomMachineType.from_str("n2d-custom-8-81920-ext")
    assert cmt.zone is None
    assert cmt.memory_mb == 81920
    assert cmt.extra_memory_used is True
    assert cmt.cpu_series is CustomMachineType.CPUSeries.N2D
    assert cmt.core_count == 8
